from __future__ import annotations

"""
Base Backends
"""

import abc
import time
import asyncio
import contextlib

from lzl.logging import logger, null_logger, Logger
from lzl.types import eproperty
from .config import BackendType, DBConfigT
from .utils import SQLAlchemyUtilities
import typing as t

if t.TYPE_CHECKING:
    from sqlalchemy import Connection, Engine
    from sqlalchemy.orm import Session, session, sessionmaker, scoped_session
    from sqlalchemy.ext.asyncio import async_sessionmaker, async_scoped_session
    from sqlalchemy.ext.asyncio import AsyncSession, AsyncEngine, AsyncConnection

    from sqlmodel import Session as SMSession
    from sqlmodel.ext.asyncio.session import AsyncSession as SMAsyncSession

    SessionT = t.Union[SMSession, Session]
    AsyncSessionT = t.Union[SMAsyncSession, AsyncSession]

    SessionMakerT = sessionmaker[SessionT]
    AsyncSessionMakerT = async_sessionmaker[AsyncSessionT]

    ScopedSessionMakerT = scoped_session[SessionT]
    AsyncScopedSessionMakerT = async_scoped_session[AsyncSessionT]

    SessionGetT = t.Union[SessionMakerT, ScopedSessionMakerT]
    AsyncSessionGetT = t.Union[AsyncSessionMakerT, AsyncScopedSessionMakerT]

    SessionGetterT = t.Union[SessionGetT, AsyncSessionGetT]

    SMSessionMakerT = sessionmaker[SMSession]
    SMAsyncSessionMakerT = async_sessionmaker[SMAsyncSession]

    SMScopedSessionMakerT = scoped_session[SMSession]
    SMAsyncScopedSessionMakerT = async_scoped_session[SMAsyncSession]

    SMSessionGetT = t.Union[SMSessionMakerT, SMScopedSessionMakerT]
    SMAsyncSessionGetT = t.Union[SMAsyncSessionMakerT, SMAsyncScopedSessionMakerT]
    SMSessionGetterT = t.Union[SMSessionGetT, SMAsyncSessionGetT]

    from lzl.pool import ThreadPool
    from sqlalchemy import TextClause


class BaseDatabaseBackend(abc.ABC):
    """
    Base Database Connection
    """
    name: t.Optional[str] = 'default'
    backend: t.Optional[BackendType] = None
    
    _extra: t.Dict[str, t.Any] = {}

    def __init__(
        self,
        config: DBConfigT,
        name: t.Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initializes the Database Connection
        """
        if name: self.name = name
        elif config.name: self.name = config.name
        self.config = config
        self.extra_kws = kwargs
        self._default_scoped = {
            'sync': self.config.session.scoped,
            'async': self.config.asession.scoped,
        }
        self.db_type = self.config.adapter_prefix
        self.db_type_name = self.db_type.capitalize()
        if not self.db_type_name.endswith('DB'): self.db_type_name += 'DB'
        self.post_init(**kwargs)

    def post_init(self, **kwargs):
        """
        Post Initialization
        """
        pass

    @eproperty
    def tpool(self) -> 'ThreadPool':
        """
        Returns the ThreadPool
        """
        from lzl.pool import ThreadPool
        return ThreadPool
    
    @property
    def url(self) -> str:
        """
        Returns the URL
        """
        return str(self.config.url)
    

    @eproperty
    def engine(self) -> 'Engine':
        """
        Returns the engine
        """
        from sqlalchemy import create_engine
        return create_engine(
            **self.config.get_engine_kwargs(
                readonly = False,
                mode = 'sync',
            )
        )
    


    @eproperty
    def engine_su(self) -> 'Engine':
        """
        Returns the superuser engine
        """
        from sqlalchemy import create_engine
        return create_engine(
            **self.config.get_engine_kwargs(
                readonly = False,
                superuser = True,
                mode = 'sync',
            )
        )
    

    @eproperty
    def engine_ro(self) -> 'Engine':
        """
        Returns the readonly engine
        """
        from sqlalchemy import create_engine
        return create_engine(
            **self.config.get_engine_kwargs(
                readonly = True,
                mode = 'sync',
            )
        )

    @eproperty
    def aengine(self) -> 'AsyncEngine':
        """
        [Async] Returns the engine
        """
        from sqlalchemy.ext.asyncio import create_async_engine
        return create_async_engine(
            **self.config.get_engine_kwargs(
                readonly = False,
                mode = 'async',
            )
        )
    

    @eproperty
    def aengine_su(self) -> 'AsyncEngine':
        """
        [Async] Returns the superuser engine
        """
        from sqlalchemy.ext.asyncio import create_async_engine
        return create_async_engine(
            **self.config.get_engine_kwargs(
                readonly = False,
                superuser = True,
                mode = 'async',
            )
        )
    

    @eproperty
    def aengine_ro(self) -> 'AsyncEngine':
        """
        [Async] Returns the readonly engine
        """
        from sqlalchemy.ext.asyncio import create_async_engine
        return create_async_engine(
            **self.config.get_engine_kwargs(
                readonly = True,
                mode = 'async',
            )
        )

    @eproperty
    def session(self) -> 'SessionMakerT':
        """
        Returns the session
        """
        from sqlalchemy.orm import sessionmaker
        return sessionmaker(
            self.engine,
            **self.config.get_session_kwargs(
                readonly = False,
                mode = 'sync',
                backend = self.backend,
            )
        )
    
    @eproperty
    def session_ro(self) -> 'SessionMakerT':
        """
        Returns the readonly session
        """
        from sqlalchemy.orm import sessionmaker
        return sessionmaker(
            self.engine_ro,
            **self.config.get_session_kwargs(
                readonly = True,
                mode = 'sync',
                backend = self.backend,
            )
        )
    
    @eproperty
    def session_su(self) -> 'SessionMakerT':
        """
        Returns the superuser session
        """
        from sqlalchemy.orm import sessionmaker
        return sessionmaker(
            self.engine_su,
            **self.config.get_session_kwargs(
                readonly = False,
                superuser = True,
                mode = 'sync',
                backend = self.backend,
            )
        )
    
    @eproperty
    def asession(self) -> 'AsyncSessionMakerT':
        """
        [Async] Returns the session
        """
        from sqlalchemy.ext.asyncio import async_sessionmaker
        return async_sessionmaker(
            self.aengine,
            **self.config.get_session_kwargs(
                readonly = False,
                mode = 'async',
                backend = self.backend,
            )
        )
    
    @eproperty
    def asession_su(self) -> 'AsyncSessionMakerT':
        """
        [Async] Returns the superuser session
        """
        from sqlalchemy.ext.asyncio import async_sessionmaker
        return async_sessionmaker(
            self.aengine_su,
            **self.config.get_session_kwargs(
                readonly = False,
                superuser = True,
                mode = 'async',
                backend = self.backend,
            )
        )

    @eproperty
    def asession_ro(self) -> 'AsyncSessionMakerT':
        """
        [Async] Returns the readonly session
        """
        from sqlalchemy.ext.asyncio import async_sessionmaker
        return async_sessionmaker(
            self.aengine_ro,
            **self.config.get_session_kwargs(
                readonly = True,
                mode = 'async',
                backend = self.backend,
            )
        )
    
    """
    Scoped Sessions
    """

    @eproperty
    def session_scoped(self) -> 'ScopedSessionMakerT':
        """
        Returns the Scoped Session 
        """
        from sqlalchemy.orm import scoped_session
        return scoped_session(
            self.session,
            scopefunc = self.extra_kws.get('scopefunc', None),
        )
    
    @eproperty
    def session_ro_scoped(self) -> 'ScopedSessionMakerT':
        """
        Returns the readonly Scoped Session
        """
        from sqlalchemy.orm import scoped_session
        return scoped_session(
            self.session_ro,
            scopefunc = self.extra_kws.get('scopefunc', None),
        )
        
    
    @eproperty
    def asession_scoped(self) -> 'AsyncScopedSessionMakerT':
        """
        [Async] Returns the Scoped Session
        """
        from sqlalchemy.ext.asyncio import async_scoped_session
        return async_scoped_session(
            self.asession,
            scopefunc = self.extra_kws.get('scopefunc', None),
        )
        
    
    @eproperty
    def asession_ro_scoped(self) -> 'AsyncScopedSessionMakerT':
        """
        [Async] Returns the readonly Scoped Session
        """
        from sqlalchemy.ext.asyncio import async_scoped_session
        return async_scoped_session(
            self.asession_ro,
            scopefunc = self.extra_kws.get('scopefunc', None),
        )
        

    @eproperty
    def logger(self) -> 'Logger':
        """
        Returns the logger
        """
        return logger
    
    @eproperty
    def autologger(self) -> 'Logger':
        """
        Returns the autologger
        """
        return logger if self.config.debug_enabled else null_logger
    

    def _get_session_type(
        self,
        readonly: t.Optional[bool] = None,
        scoped: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
    ) -> t.Tuple['SessionGetT', bool]:
        """
        Returns the session
        """
        if superuser: return (self.session_su, False)
        if scoped is None: scoped = self._default_scoped['sync']
        if readonly: return (self.session_ro_scoped, True) if scoped else (self.session_ro, False)
        return (self.session_scoped, True) if scoped else (self.session, False)
    
    def _get_async_session_type(
        self,
        readonly: t.Optional[bool] = None,
        scoped: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
    ) -> t.Tuple['AsyncSessionGetT', bool]:
        """
        Returns the session
        """
        if superuser: return (self.asession_su, False)
        if scoped is None: scoped = self._default_scoped['async']
        if readonly: return (self.asession_ro_scoped, True) if scoped else (self.asession_ro, False)
        return (self.asession_scoped, True) if scoped else (self.asession, False)


    def _get_session(
        self,
        readonly: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        mode: t.Optional[t.Literal['sync', 'async']] = 'sync',
        scoped: t.Optional[bool] = None,
    ) -> t.Tuple['SessionGetterT', bool]:
        """
        Returns the session
        """
        if mode == 'sync': return self._get_session_type(readonly = readonly, scoped = scoped, superuser = superuser)
        return self._get_async_session_type(readonly = readonly, scoped = scoped, superuser = superuser)


    """
    Direct Engine Connections
    """

    def get_engine(
        self,
        url: t.Optional[str] = None,
        readonly: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        execution_options: t.Optional[t.Dict[str, t.Any]] = None,
        adapter: t.Optional[str] = None,
        **kwargs,
    ) -> 'Engine':
        """
        [Sync] Creates an Engine. This is unmanaged
        """
        from sqlalchemy import create_engine
        engine_kwargs = self.config.get_engine_kwargs(
            readonly = readonly, 
            superuser = superuser, 
            url = url, 
            adapter = adapter,
            mode = 'sync',
            **kwargs
        )
        if execution_options: engine_kwargs['execution_options'] = execution_options
        return create_engine(**engine_kwargs)
    
    def get_aengine(
        self,
        url: t.Optional[str] = None,
        readonly: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        execution_options: t.Optional[t.Dict[str, t.Any]] = None,
        adapter: t.Optional[str] = None,
        **kwargs,
    ) -> 'AsyncEngine':
        """
        [Async] Creates an Async Engine. This is unmanaged
        """
        from sqlalchemy.ext.asyncio import create_async_engine
        engine_kwargs = self.config.get_engine_kwargs(
            readonly = readonly, 
            superuser = superuser, 
            url = url, 
            adapter = adapter,
            mode = 'async',
            **kwargs
        )
        if execution_options: engine_kwargs['execution_options'] = execution_options
        return create_async_engine(**engine_kwargs)
    
    """
    Raw Connection Methods
    """

    def _get_engine_type(
        self,
        readonly: t.Optional[bool] = None,
        mode: t.Optional[t.Literal['sync', 'async']] = 'sync',
    ) -> t.Union['Engine', 'AsyncEngine']:
        """
        Returns the session type
        """
        if mode == 'sync': return self.engine_ro if readonly else self.engine
        return self.aengine_ro if readonly else self.aengine
    

    @contextlib.contextmanager
    def connection(
        self,
        readonly: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        execution_options: t.Optional[t.Dict[str, t.Any]] = None,
        url: t.Optional[str] = None,
        **kwargs,
    ) -> t.Generator['Connection', None, None]:
        """
        Context Manager for the connection
        """
        if any(
            url is not None,
            execution_options is not None,
            superuser is not None,
        ):
            engine = self.get_engine(
                url = url,
                readonly = readonly,
                superuser = superuser,
                execution_options = execution_options,
                **kwargs,
            )
        else:
            engine = self._get_engine_type(readonly = readonly)
        conn: Connection = None
        try:
            conn = engine.connect()
            conn.begin()
            yield conn
        except Exception as e:
            self.logger.trace(f'[{self.name} - {self.backend}] Engine Error', e, depth = 5)
            if auto_rollback is None: auto_rollback = self.config.auto_rollback_enabled
            if auto_rollback: 
                try:
                    conn.rollback()
                    self.logger.info(f'[{self.name} - {self.backend}] Connection Rolled Back')
                except Exception as e:
                    self.logger.trace(f'[{self.name} - {self.backend}] Connection Rollback Error', e)
            if raise_errors: raise e
        finally:
            if auto_commit is None: auto_commit = self.config.autocommit_enabled
            if auto_commit: conn.commit()
            conn.close()

    @contextlib.asynccontextmanager
    async def aconnection(
        self,
        readonly: t.Optional[bool] = None,
        raise_errors: t.Optional[bool] = None,
        auto_rollback: t.Optional[bool] = None,
        auto_commit: t.Optional[bool] = None,
        superuser: t.Optional[bool] = None,
        execution_options: t.Optional[t.Dict[str, t.Any]] = None,
        url: t.Optional[str] = None,
        **kwargs,
    ) -> t.AsyncGenerator['AsyncConnection', None]:
        """
        Context Manager for the connection
        """
        if any(
            url is not None,
            execution_options is not None,
            superuser is not None,
        ):
            engine = self.get_aengine(
                url = url,
                readonly = readonly,
                superuser = superuser,
                execution_options = execution_options,
                **kwargs,
            )
        else:
            engine = self._get_engine_type(readonly = readonly, mode = 'async')
        conn: AsyncConnection = None
        try:
            conn = await engine.connect()
            await conn.start()
            yield conn
        except Exception as e:
            self.logger.trace(f'[{self.name} - {self.backend}] Engine Error', e, depth = 5)
            if auto_rollback is None: auto_rollback = self.config.auto_rollback_enabled
            if auto_rollback: 
                try:
                    await conn.rollback()
                    logger.info('Connection Rolled Back')
                except Exception as e:
                    logger.trace('Connection Rollback Error', e)
            if raise_errors: raise e
        finally:
            if auto_commit is None: auto_commit = self.config.autocommit_enabled
            if auto_commit: await conn.commit()
            await conn.aclose()



    @eproperty
    def sql(self) -> 'SQLAlchemyUtilities':
        """
        Returns the SQLAlchemy Utilities
        """
        return SQLAlchemyUtilities()

    """
    Utility Methods
    """


    def dbexec(
        self, 
        sql: str,
        superuser: t.Optional[bool] = None,
        readonly: t.Optional[bool] = None,
        database: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        verbose: t.Optional[bool] = True,
        **kwargs,
    ) -> str:
        """
        Runs a Raw SQL Query through the Postgres CLI
        """
        if not sql.endswith(';'): sql += ';'
        cmd = f'{self.config.get_cli_connect_string(superuser=superuser, readonly=readonly, database=database)} -c "{sql}"'
        # if debug_enabled: cmd += ' -v ON_ERROR_STOP=1 -a -b -e -E'
        # safe_cmd = cmd.split("psql", 1)[-1]
        if debug_enabled: cmd += self.config.get_cli_debug_args()
        safe_cmd = self.config.get_cli_safe_cmd(cmd)
        if verbose: self.logger.info(f'Executing Command: {safe_cmd}', prefix = f'{self.name} - {self.backend}', colored = True)        
        import subprocess
        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result = ''
        for line in process.stdout:
            if isinstance(line, bytes): line = line.decode('utf-8')
            result += f'{line.strip()}\n'
        if verbose: self.logger.info(f'Result: {result}', prefix = f'{self.name} - {self.backend}', colored = True)
        return result

    async def adbexec(
        self, 
        sql: str,
        superuser: t.Optional[bool] = None,
        readonly: t.Optional[bool] = None,
        database: t.Optional[str] = None,
        debug_enabled: t.Optional[bool] = None,
        verbose: t.Optional[bool] = True,
        **kwargs,
    ) -> str:
        """
        Runs a Raw SQL Query through the Postgres CLI
        """
        if not sql.endswith(';'): sql += ';'
        cmd = f'{self.config.get_cli_connect_string(superuser=superuser, readonly=readonly, database=database)} -c "{sql}"'
        # if debug_enabled: cmd += ' -v ON_ERROR_STOP=1 -a -b -e -E'
        if debug_enabled: cmd += self.config.get_cli_debug_args()
        safe_cmd = self.config.get_cli_safe_cmd(cmd)
        # safe_cmd = cmd.split("psql", 1)[-1]
        if verbose: self.logger.info(f'Executing Command: {safe_cmd}', prefix = f'{self.name} - {self.backend}', colored = True)
        process = await self.tpool.async_run_command(cmd, output_only=False)
        result = ''
        async for line in process.stdout:
            if isinstance(line, bytes): line = line.decode('utf-8')
            result += f'{line.strip()}\n'
        if verbose: self.logger.info(f'Result: {result}', prefix = f'{self.name} - {self.backend}', colored = True)
        return result


    def wait_for_ready(
        self,
        max_attempts: t.Optional[int] = 10,
        interval: t.Optional[float] = 5.0,
    ):
        """
        Waits for the database to be ready
        """
        attempts = 0
        while attempts < max_attempts:
            try:
                with self.engine.begin() as conn:
                    conn.execute(self.sql.text('SELECT 1'))
                return
            except Exception as e:
                if attempts == 0:
                    logger.info(f'Database not ready: {attempts}/{max_attempts} attempts @ {self.config.safe_url}. Error: {e}', prefix = f'{self.name} - {self.backend}')
                else:    
                    logger.info(f'Database not ready: {attempts}/{max_attempts} attempts. Error: {e}', prefix = f'{self.name} - {self.backend}')
                attempts += 1
                time.sleep(interval)
                if attempts >= max_attempts:
                    raise e

    async def await_for_ready(
        self,
        max_attempts: t.Optional[int] = 10,
        interval: t.Optional[float] = 5.0,
    ):
        """
        Waits for the database to be ready
        """
        attempts = 0
        while attempts < max_attempts:
            try:
                async with self.aengine.begin() as conn:
                    await conn.execute(self.sql.text('SELECT 1'))
                return
            except Exception as e:
                if attempts == 0:
                    logger.info(f'Database not ready: {attempts}/{max_attempts} attempts @ {self.config.safe_url}. Error: {e}', prefix = f'{self.name} - {self.backend}')
                else:    
                    logger.info(f'Database not ready: {attempts}/{max_attempts} attempts. Error: {e}', prefix = f'{self.name} - {self.backend}')
                attempts += 1
                await asyncio.sleep(interval)
                if attempts >= max_attempts:
                    raise e
    
    def _execute_scalar(
        self,
        query: 'TextClause',
        params: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> t.Any:
        """
        Executes a scalar query
        """
        with self.session() as session:
            result = session.execute(query, params = params, **kwargs)
            return result.scalar_one()
    
    async def _aexecute_scalar(
        self,
        query: 'TextClause',
        params: t.Optional[t.Dict[str, t.Any]] = None,
        **kwargs,
    ) -> t.Any:
        """
        Executes a scalar query
        """
        async with self.asession() as session:
            result = await session.execute(query, params = params, **kwargs)
            return result.scalar_one()

    def index_exists(
        self,
        index_name: str,
    ) -> bool:
        """
        Checks if the Index Exists
        """
        statement = self.sql.text(
            """
            SELECT EXISTS (
                SELECT FROM pg_indexes
                WHERE indexname = :index_name
            );
            """
        )
        return self._execute_scalar(statement, {'index_name': index_name})

    
    async def aindex_exists(
        self,
        index_name: str,
    ) -> bool:
        """
        Checks if the Index Exists
        """
        statement = self.sql.text(
            """
            SELECT EXISTS (
                SELECT FROM pg_indexes
                WHERE indexname = :index_name
            );
            """
        )
        return await self._aexecute_scalar(statement, {'index_name': index_name})


    def database_exists(
        self,
        db_name: str,
        verbose: t.Optional[bool] = True,
    ) -> bool:
        """
        Checks if the database exists
        """
        statement = self.sql.text(
            f"SELECT 1 FROM pg_database WHERE datname = '{db_name}';"
        )
        result = self._execute_scalar(statement)
        if verbose: logger.info(f'Database {db_name} Exists: {result}')
        return result
        

    async def adatabase_exists(
        self,
        db_name: str,
        verbose: t.Optional[bool] = True,
    ) -> bool:
        """
        Checks if the database exists
        """
        statement = self.sql.text(
            f"SELECT 1 FROM pg_database WHERE datname = '{db_name}';"
        )
        result = await self._aexecute_scalar(statement)
        if verbose: logger.info(f'Database {db_name} Exists: {result}')
        return result
    

    def get_table_names(
        self,
        schema: t.Optional[str] = None,
    ) -> t.List[str]:
        """
        Returns the table names
        """
        from sqlalchemy.engine.reflection import Inspector
        inspector = Inspector.from_engine(self.engine)
        return inspector.get_table_names(schema = schema)
    
    def get_table_column_size(
        self,
        table: str,
        column: str,
        schema: t.Optional[str] = None,
        verbose: t.Optional[bool] = False,
    ) -> t.Dict[str, t.Union[str, float]]:
        """
        Displays the table column size
        """
        table_name = f'{schema}.{table}' if schema else table
        with self.connection(readonly=True) as conn:
            result = conn.execute(
                self.sql.text(f'''
                select
                    pg_size_pretty(sum(pg_column_size({column}))) as total_size,
                    pg_size_pretty(avg(pg_column_size({column}))) as average_size,
                    sum(pg_column_size({column})) * 100.0 / pg_total_relation_size('{table_name}') as percentage
                from {table_name};
                '''
            ))
        row = result.fetchone()
        if verbose: self.logger.info(f'Total Size: |g|{row[0]}|e|, Avg Size: |g|{row[1]}|e|, Percentage: |y|{row[2]:.2f}%|e|', prefix = f'{table_name}.{column}', colored = True)
        else: self.autologger.info(f'Total Size: |g|{row[0]}|e|, Avg Size: |g|{row[1]}|e|, Percentage: |y|{row[2]:.2f}%|e|', prefix = f'{table_name}.{column}', colored = True)
        return {
            'total_size': row[0],
            'average_size': row[1],
            'percentage': row[2],
        }
    
    def get_table_column_names(
        self,
        table: str,
        schema: t.Optional[str] = None,
    ) -> t.List[str]:
        """
        Returns the table column names
        """
        table_name = f'{schema}.{table}' if schema else table
        with self.connection(readonly=True) as conn:
            result = conn.execute(
                self.sql.text(f'''
                select column_name
                from information_schema.columns
                where table_name = '{table_name}';
                '''
            ))
        return [row[0] for row in result.fetchall()]


    def upgrade_extensions(
        self,
        extensions: t.Optional[t.List[str]] = None,
        verbose: t.Optional[bool] = True,
        **kwargs,
    ) -> t.Dict[str, t.Any]:
        """
        Upgrade the extensions
        """
        if not extensions: return
        results = {}
        with self.engine.begin() as conn:
            for ext in extensions:
                conn.execute(self.sql.text(f"CREATE EXTENSION IF NOT EXISTS {ext}"))
                conn.commit()
                conn.execute(self.sql.text(f"ALTER EXTENSION {ext} UPDATE"))
                conn.commit()

                # Get the version
                result = conn.execute(self.sql.text(f"SELECT * FROM pg_extension WHERE extname = '{ext}'"))
                version = result.scalar()
                results[ext] = version
                if verbose: self.logger.info(f"Extension {ext} version: {version}")
                else: self.autologger.info(f"Extension {ext} version: {version}")
        return results


    async def aupgrade_extensions(
        self,
        extensions: t.Optional[t.List[str]] = None,
        verbose: t.Optional[bool] = True,
        **kwargs,
    ) -> t.Dict[str, t.Any]:
        """
        Upgrade the extensions
        """
        if not extensions: return
        results = {}
        async with self.aengine.begin() as conn:
            for ext in extensions:
                await conn.execute(self.sql.text(f"CREATE EXTENSION IF NOT EXISTS {ext}"))
                await conn.commit()
                await conn.execute(self.sql.text(f"ALTER EXTENSION {ext} UPDATE"))
                await conn.commit()

                # Get the version
                result = await conn.execute(self.sql.text(f"SELECT * FROM pg_extension WHERE extname = '{ext}'"))
                version = await result.scalar()
                results[ext] = version
                if verbose: self.logger.info(f"Extension {ext} version: {version}")
                else: self.autologger.info(f"Extension {ext} version: {version}")
        return results



BackendClassT = t.TypeVar('BackendClassT', bound = BaseDatabaseBackend)



from __future__ import annotations

"""
Temporal Workflow Mixins: Dispatch
"""

import typing as t
from lzl.types import eproperty

from .base import BaseTemporalMixin, MixinKinds

if t.TYPE_CHECKING:
    from temporalio.types import ParamType, ReturnType, SelfType
    from lzl.ext.temporal.patches import WorkflowHandle
    # from temporalio.client import WorkflowHandle


ParamT = t.TypeVar('ParamT')
ReturnT = t.TypeVar('ReturnT')

class TemporalDispatchMixin(BaseTemporalMixin, t.Generic[ParamT, ReturnT]):
    """
    [Temporal] Dispatch Mixin that will automatically registered
    """

    workflow_name: t.Optional[str] = None
    mixin_kind: t.Optional[MixinKinds] = 'dispatch'
    _is_subclass_: t.Optional[bool] = True

    input_type: t.Optional[t.Type[ParamT] | str] = None
    result_type: t.Optional[t.Type[ReturnT] | str] = None

    # task_queue: t.Optional[str] = 'default'

    @eproperty
    def workflow_func_name(self) -> str:
        """
        Returns the workflow function name
        `task.function` -> `task_function`
        """
        return self.workflow_name.replace('.', '_').replace('-', '_')
            
    
    async def _precall_hook_(self, *args, **kwargs):
        """
        Runs the dispatch hook
        """
        if self.client is None: 
            self.client = await self.registry.get_client(namespace = self.namespace)
        if not hasattr(self, 'client_func_kws'):
            self.client_func_kws = {
                'execute_workflow': list(
                    self.client.execute_workflow.__code__.co_varnames
                ),
                'start_workflow': list(
                    self.client.start_workflow.__code__.co_varnames
                ),
                'create_schedule': list(
                    self.client.create_schedule.__code__.co_varnames
                ),
            }

    def _extract_func_kwargs_(self, func: str, kwargs: t.Dict[str, t.Any]) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        """
        Extracts the function kwargs
        """
        func_kws = self.client_func_kws[func]
        return {
            k: v for k, v in kwargs.items() if k not in func_kws
        }, {
            k: v for k, v in kwargs.items() if k in func_kws
        }
        
    async def _validate_input_(self, param: 'ParamT', kwargs: t.Dict[str, t.Any]) -> 'ParamT':
        """
        Validates the input

        - if Arguments are provided, it will return the first argument
        - if no arguments are provided, it will return the kwargs
        """
        p = param[0] if param else kwargs
        return self.input_type(**p) if isinstance(p, dict) and self.input_type else p

    async def _validate_output_(self, result: 'ReturnType') -> 'ReturnType':
        """
        Validates the output
        """
        if isinstance(result, dict) and self.result_type:
            return self.result_type(**result)
        return result
    
    async def _finalize_params_(self, params: 'ParamT', id: str, kwargs: t.Dict[str, t.Any]) -> 'ParamT':
        """
        Finalizes the params
        """
        return params
    
    async def _execute_(
        self, 
        *param: 'ParamT',
        id: t.Optional[str] = None,
        task_queue: t.Optional[str] = None,
        workflow: t.Optional[str] = None,
        **kwargs,
    ) -> 'ReturnT':  # sourcery skip: avoid-builtin-shadow
        """
        Executes the Workflow
        """
        await self._precall_hook_()
        params, kwargs = self._extract_func_kwargs_('execute_workflow', kwargs)
        params = await self._validate_input_(param, params)
        if workflow is None: workflow = self.workflow_name
        if task_queue is None: task_queue = self.task_queue
        if id is None: id = self.generate_id(params)
        params = await self._finalize_params_(params, id, kwargs)
        result = await self.client.execute_workflow(
            workflow,
            params,
            id = id,
            task_queue = task_queue,
            **kwargs,
        )
        return await self._validate_output_(result)

    async def _execute_async_(
        self, 
        *param: 'ParamT',
        id: t.Optional[str] = None,
        task_queue: t.Optional[str] = None,
        workflow: t.Optional[str] = None,
        **kwargs,
    ) -> 'WorkflowHandle[SelfType, ReturnT]':  # sourcery skip: avoid-builtin-shadow
        """
        Executes the Workflow Asyncronously and Returns the Workflow Handle
        """
        await self._precall_hook_()
        params, kwargs = self._extract_func_kwargs_('start_workflow', kwargs)
        params = await self._validate_input_(param, params)
        # print(f'params: {params} ({type(params)}), kwargs: {kwargs}')
        if workflow is None: workflow = self.workflow_name
        if task_queue is None: task_queue = self.task_queue
        if id is None: id = self.generate_id(params)
        params = await self._finalize_params_(params, id, kwargs)
        wf: WorkflowHandle = await self.client.start_workflow(
            workflow,
            params,
            id = id,
            task_queue = task_queue,
            **kwargs,
        )
        wf.set_validator(self._validate_output_)
        return wf

    async def run(
        self, 
        *param: 'ParamT',
        id: t.Optional[str] = None,
        task_queue: t.Optional[str] = None,
        workflow: t.Optional[str] = None,
        **kwargs,
    ) -> 'ReturnT':  # sourcery skip: avoid-builtin-shadow
        """
        Executes the Workflow
        """
        return await self._execute_(*param, id = id, task_queue = task_queue, workflow = workflow, **kwargs)

    async def arun(
        self, 
        *param: 'ParamT',
        id: t.Optional[str] = None,
        task_queue: t.Optional[str] = None,
        workflow: t.Optional[str] = None,
        **kwargs,
    ) -> 'WorkflowHandle[SelfType, ReturnT]':  # sourcery skip: avoid-builtin-shadow
        """
        Executes the Workflow Asyncronously and Returns the Workflow Handle
        """
        return await self._execute_async_(*param, id = id, task_queue = task_queue, workflow = workflow, **kwargs)

    @classmethod
    def _on_preregister_hook_(cls):
        """
        Runs the dispatch pre-register hook
        """
        pass

    @classmethod
    def _on_register_hook_(cls):
        """
        Runs the dispatch hook
        """
        pass

    def _on_init_hook_(self):
        """
        Runs the dispatch init hook
        """
        pass
    



from __future__ import annotations

"""No-op logger implementation that still triggers LazyOps hooks."""

import logging
import traceback
import typing as t

from .mixins import LoggingMixin
from .utils import format_item, format_message, get_logging_level

if t.TYPE_CHECKING:
    from .base import Logger
    from .utils import MsgItem


class NullLoggerV1(logging.Logger):
    """Backwards-compatible stand-in that simply swallows log calls."""

    def info(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def debug(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def warning(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def error(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def critical(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def exception(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def log(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def trace(self, *args: t.Any, **kwargs: t.Any) -> None: ...
    def success(self, *args: t.Any, **kwargs: t.Any) -> None: ...


class NullLogger(logging.Logger, LoggingMixin):
    """Logger that delegates output to hooks without emitting messages."""

    def _format_item(
        self,
        msg: MsgItem,
        max_length: int | None = None,
        colored: bool | None = False,
        level: str | None = None,
        _is_part: bool | None = False,
    ) -> str:
        return format_item(msg, max_length=max_length, colored=colored, level=level, _is_part=_is_part)

    def _format_message(
        self,
        message: MsgItem,
        *args: MsgItem,
        prefix: str | None = None,
        max_length: int | None = None,
        level: str | None = None,
        colored: bool | None = False,
        extra: dict[str, t.Any] | None = None,
    ) -> str:
        return format_message(
            message,
            *args,
            prefix=prefix,
            max_length=max_length,
            level=level,
            colored=colored,
            extra=extra,
        )

    def _get_level(self, level: t.Union[str, int]) -> str:
        return get_logging_level(level)

    def log(
        self,
        level: t.Union[str, int],
        message: t.Any,
        *args: MsgItem,
        prefix: str | None = None,
        max_length: int | None = None,
        colored: bool | None = False,
        hook: t.Callable[[str], None] | None = None,
        **kwargs: t.Any,
    ) -> None:  # noqa: N805
        if not hook:
            return
        resolved_level = self._get_level(level)
        extra = kwargs.get("extra")
        rendered = self._format_message(
            message,
            *args,
            prefix=prefix,
            max_length=max_length,
            colored=colored,
            level=resolved_level,
            extra=extra,
        )
        self.run_logging_hooks(rendered, hook=hook)

    def info(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.log("INFO", *args, **kwargs)

    def debug(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.log("DEBUG", *args, **kwargs)

    def warning(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.log("WARNING", *args, **kwargs)

    def error(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.log("ERROR", *args, **kwargs)

    def trace(
        self,
        msg: t.Union[str, t.Any],
        error: type[BaseException] | None = None,
        level: str = "ERROR",
        limit: int | None = None,
        chain: bool | None = True,
        colored: bool | None = False,
        prefix: str | None = None,
        max_length: int | None = None,
        hook: t.Callable[[str], None] | None = None,
        **kwargs: t.Any,
    ) -> None:
        if not hook:
            return
        depth = kwargs.pop("depth", None)
        if depth is not None:
            limit = depth
        extra = kwargs.get("extra")
        if isinstance(msg, str):
            rendered = msg
            if extra:
                extra_rendered = format_item(extra, max_length=max_length, colored=colored, level=level)
                extra_rendered = extra_rendered.lstrip("\n")
                if extra_rendered:
                    rendered = f"{rendered}\n{extra_rendered}" if rendered else extra_rendered
        else:
            rendered = self._format_message(
                msg,
                colored=colored,
                prefix=prefix,
                max_length=max_length,
                level=level,
                extra=extra,
            )
        rendered += f"\n{traceback.format_exc(chain=chain, limit=limit)}"
        if error:
            rendered += f" - {error}"
        self.run_logging_hooks(rendered, hook=hook)

    def exception(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.log("ERROR", *args, **kwargs)

    def success(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.log("SUCCESS", *args, **kwargs)


__all__ = ["NullLogger", "NullLoggerV1"]



from __future__ import annotations


from enum import Enum
from .static import DEFAULT_CLASS_COLOR, DEFAULT_FUNCTION_COLOR, RESET_COLOR, QUEUE_STATUS_COLORS, FALLBACK_STATUS_COLOR
from typing import Dict, Any, Union


class LoggerFormatter:

    max_extra_lengths: Dict[str, int] = {}

    @classmethod
    def get_extra_length(cls, key: str, value: str) -> int:
        """
        Returns the max length of an extra key
        """
        if key not in cls.max_extra_lengths:
            cls.max_extra_lengths[key] = len(key)
        if len(value) > cls.max_extra_lengths[key]:
            cls.max_extra_lengths[key] = len(value)
        return cls.max_extra_lengths[key]

    @classmethod
    def queue_logger_formatter(cls, record: Dict[str, Union[Dict[str, Any], Any]]) -> str:
        """
        Formats the log message for the queue.
        """
        _extra: Dict[str, Union[Dict[str, Any], Any]] = record.get('extra', {})
        if not record['extra'].get('worker_name'):
            record['extra']['worker_name'] = ''
        
        status = _extra.get('status')
        kind: str = _extra.get('kind')
        if status and isinstance(status, Enum): status = status.name
        
        kind_color = QUEUE_STATUS_COLORS.get(kind.lower(), FALLBACK_STATUS_COLOR)
        if '<' not in kind_color: kind_color = f'<{kind_color}>'
        extra = kind_color + '{extra[kind]}</>:'
        if _extra.get('queue_name'):
            queue_name_length = cls.get_extra_length('queue_name', _extra['queue_name'])
            extra += '<b><fg #006d77>{extra[queue_name]:<' + str(queue_name_length) + '}</></>:'
        if _extra.get('worker_name'):
            worker_name_length = cls.get_extra_length('worker_name', _extra['worker_name'])
            extra += '<fg #83c5be>{extra[worker_name]:<' + str(worker_name_length) + '}</>:'
        # extra += '<fg #83c5be>{extra[worker_name]}</>:<b><fg #006d77>{extra[queue_name]:<18}</></>:'
        if _extra.get('job_id'):
            extra += '<fg #005f73>{extra[job_id]}</>'
        if status:
            status_color = QUEUE_STATUS_COLORS.get(status.lower(), FALLBACK_STATUS_COLOR)
            if '<' not in status_color: status_color = f'<{status_color}>'
            extra += f':{status_color}' + '{extra[status]}</>: '
        # extra += RESET_COLOR
        # print(extra)
        return extra


    @classmethod
    def default_formatter(cls, record: Dict[str, Union[Dict[str, Any], Any]]) -> str:
        """
        To add a custom format for a module, add another `elif` clause with code to determine `extra` and `level`.

        From that module and all submodules, call logger with `logger.bind(foo='bar').info(msg)`.
        Then you can access it with `record['extra'].get('foo')`.
        """        
        _extra = record.get('extra', {})
        if _extra.get('module_name'):
            extra = DEFAULT_CLASS_COLOR + '{extra[module_name]}</>:' + DEFAULT_FUNCTION_COLOR + '{function}</>: '
        else:
            extra = DEFAULT_CLASS_COLOR + '{name}</>:' + DEFAULT_FUNCTION_COLOR + '{function}</>: '
        if _extra.get('worker_name') or _extra.get('queue_name'):
            extra = cls.queue_logger_formatter(record)
        
        if 'result=tensor([' not in str(record['message']):
            return "<level>{level: <8}</> <green>{time:YYYY-MM-DD HH:mm:ss.SSS}</>: " \
                       + extra + "<level>{message}</level>" + RESET_COLOR + "\n"
        
        msg = str(record['message'])[:100].replace('{', '(').replace('}', ')')
        return "<level>{level: <8}</> <green>{time:YYYY-MM-DD HH:mm:ss.SSS}</>: "\
                   + extra + "<level>" + msg + f"</level>{RESET_COLOR}\n"




"""Registry utilities shared across LazyOps modules."""

from .base import MRegistry, combine_parts
from .clients import (
    get_app_client,
    register_app_client,
    register_app_clients,
    register_client,
    register_client_hook,
    register_initialized_client,
)
from .settings import (
    get_app_settings,
    register_app_settings,
    register_initialized_settings,
    register_settings,
)

__all__ = [
    'MRegistry',
    'combine_parts',
    'get_app_client',
    'get_app_settings',
    'register_app_client',
    'register_app_clients',
    'register_app_settings',
    'register_client',
    'register_client_hook',
    'register_initialized_client',
    'register_initialized_settings',
    'register_settings',
]



"""
Common Types
"""

import os
import sys
from enum import Enum, EnumMeta
from functools import singledispatchmethod
from typing import Any, Union, Dict, Optional, TypeVar, TYPE_CHECKING

if sys.version_info >= (3, 8):
    from typing import Final, Literal
else:
    from typing_extensions import Final, Literal

try:
    from aenum import extend_enum
    _EXTEND_SUPPORTED = True
except ImportError:
    _EXTEND_SUPPORTED = False

class StrEnumMeta(EnumMeta):

    @singledispatchmethod
    def __getitem__(self, key):
        return super().__getitem__(key)

    @__getitem__.register
    def _(self, index: int):
        return list(self)[index]
    
    @property
    def _reversed_members_(cls) -> Dict[str, str]:
        """
        Returns the reversed members
        """
        return {v: k for k, v in cls.__members__.items()}

from .base import PYDANTIC_VERSION

def _get_serialized_value(value: Union[str, 'StrEnum']) -> str:
    """
    Returns the serialized value
    """
    return value.name if hasattr(value, 'name') else value
    
class StrEnum(str, Enum, metaclass=StrEnumMeta):
    """
    StrEnum is a string enum that allows for case-insensitive comparisons
    """

    def __eq__(self, other: Any) -> bool:
        return self.value.lower() == other.lower() if \
            isinstance(other, str) else \
                super().__eq__(other)
    
    def __ne__(self, other: Any) -> bool:
        return self.value.lower() != other.lower() if \
            isinstance(other, str) else \
                super().__ne__(other)

    def __str__(self) -> str:
        return str.__str__(self)
    
    def __hash__(self) -> int:
        return id(self)
    
    @classmethod
    def extend(cls, name: str, value: Any):
        """
        Extends the enum with a new value
        """
        if not _EXTEND_SUPPORTED: raise ImportError('aenum is not installed. Please install it to use this feature')
        extend_enum(cls, name, value)

    @classmethod
    def __validate__(cls, value: Union[str, 'StrEnumT']) -> 'StrEnumT':
        """
        Validates the value
        """
        if hasattr(value, '__members__'): return value
        # return cls(value.__name__)
        reversed_members = {v: k for k, v in cls.__members__.items()}
        if value in cls.__members__:
            return cls.__members__[value]
        elif value in reversed_members:
            return cls.__members__[reversed_members[value]]
        elif value.lower() in cls.__members__:
            return cls.__members__[value.lower()]
        elif value.capitalize() in cls.__members__:
            return cls.__members__[value.capitalize()]
        elif value.upper() in cls.__members__:
            return cls.__members__[value.upper()]
        elif value.lower() in reversed_members:
            return cls.__members__[reversed_members[value.lower()]]
        raise ValueError(f"Invalid {cls.__name__} value: {value}")
        
    if TYPE_CHECKING:
        from pydantic_core import core_schema, SchemaSerializer
        from pydantic.annotated_handlers import GetCoreSchemaHandler, GetJsonSchemaHandler
        from pydantic.json_schema import JsonSchemaValue
    
    if PYDANTIC_VERSION == 2:
    
        @classmethod
        def __get_pydantic_json_schema__(
            cls, 
            _core_schema: 'core_schema.CoreSchema', 
            _handler: 'GetJsonSchemaHandler'
        ) -> 'JsonSchemaValue':
            """
            Get the Pydantic JSON Schema for the given source
            """
            return {'enum': [m.name for m in cls], 'type': 'string'}

        @classmethod
        def __get_pydantic_core_schema__(
            cls, 
            source: type[Any], 
            handler: 'GetCoreSchemaHandler'
        ) -> 'core_schema.CoreSchema':
            """
            Get the Pydantic CoreSchema for the given source
            """
            from pydantic_core import core_schema, SchemaSerializer
            schema = core_schema.no_info_after_validator_function(
                cls.__validate__,
                core_schema.any_schema(),
                serialization = core_schema.plain_serializer_function_ser_schema(_get_serialized_value),
                # serialization = core_schema.plain_serializer_function_ser_schema(lambda x: x),
            )
            cls.__pydantic_serializer__ = SchemaSerializer(schema)
            return schema

StrEnumT = TypeVar('StrEnumT', bound = StrEnum)
    

class UpperStrEnum(StrEnum):
    """
    UpperStrEnum is a string enum that allows for case-insensitive comparisons
    """

    def __eq__(self, other: Any) -> bool:
        return self.value.upper() == other.upper() if \
            isinstance(other, str) else \
                super().__eq__(other)
    
    def __ne__(self, other: Any) -> bool:
        return self.value.upper() != other.upper() if \
            isinstance(other, str) else \
                super().__ne__(other)
    

    def __str__(self) -> str:
        return str.__str__(self)
    
    def __hash__(self) -> int:
        return id(self)



class AppEnv(str, Enum):
    CICD = "cicd"
    DEVELOPMENT = "development"
    STAGING = "staging"
    PRODUCTION = "production"
    LOCAL = "local"
    TEST = "test"

    @classmethod
    def from_env(cls, env_value: str) -> "AppEnv":
        """
        Get the app environment from the environment variables
        """
        env_value = env_value.lower()
        if "cicd" in env_value or "ci/cd" in env_value: return cls.CICD
        if "prod" in env_value: return cls.PRODUCTION
        if "dev" in env_value: return cls.DEVELOPMENT
        if "staging" in env_value: return cls.STAGING
        if "local" in env_value: return cls.LOCAL
        if "test" in env_value: return cls.TEST
        raise ValueError(f"Invalid app environment: {env_value} ({type(env_value)})")
    
    @classmethod
    def from_hostname(cls, hostname: str) -> "AppEnv":
        """
        Get the app environment from the hostname
        """
        hostname = hostname.lower()
        if "dev" in hostname: return cls.DEVELOPMENT
        if "staging" in hostname: return cls.STAGING
        if "test" in hostname: return cls.TEST
        return cls.LOCAL if "local" in hostname else cls.PRODUCTION
        
    
    @classmethod
    def from_module_name(cls, module_name: str) -> 'AppEnv':
        """
        Retrieves the app environment
        """
        module_name = module_name.replace(".", "_").upper()
        for key in {
            "SERVER_ENV",
            f"{module_name}_ENV",
            "APP_ENV",
            "ENVIRONMENT",
        }:
            if env_value := os.getenv(key):
                return cls.from_env(env_value)

        from lzo.utils.system import is_in_kubernetes, get_host_name
        if is_in_kubernetes():
            hn = get_host_name()
            try:
                parts = hn.split("-")
                for p in parts:
                    if all(
                        e not in p.lower()
                        for e in {'development', 'test', 'staging', 'local', 'dev', 'prod', 'production'}
                    ):
                        parts.remove(p)
                return cls.from_env(parts[0]) if len(parts) > 0 else cls.PRODUCTION
                # return cls.PRODUCTION
                # return cls.from_env(parts[2]) if len(parts) > 3 else cls.PRODUCTION
            except Exception as e:
                return cls.from_hostname(hn)

        return cls.LOCAL
    
    def __eq__(self, other: Any) -> bool:
        """
        Equality operator
        """
        if isinstance(other, str): return self.value == other.lower()
        return self.value == other.value if isinstance(other, AppEnv) else False

    @property
    def is_devel(self) -> bool:
        """
        Returns True if the app environment is development
        """
        return self in [self.LOCAL, self.CICD, self.DEVELOPMENT, self.STAGING, self.TEST]

    @property
    def is_local(self) -> bool:
        """
        Returns True if the app environment is local
        """
        return self in [self.LOCAL, self.CICD]

    @property
    def name(self) -> str:
        """
        Returns the name in lower
        """
        return self.value.lower()

    @property
    def short_name(self) -> str:
        """
        Returns the short name in lower
        """
        if self == self.DEVELOPMENT: return 'dev'
        return 'prod' if self == self.PRODUCTION else self.name
    
    def select(self, values: Dict[Union[str, 'AppEnv'], Any], default: Optional[Any] = None) -> Any:
        """
        Returns the value for the app env
        """
        return next((value for key, value in values.items() if key == self), default)


    @classmethod
    def extend(cls, name: str, value: Any):
        """
        Extends the enum with a new value
        """
        if not _EXTEND_SUPPORTED: raise ImportError('aenum is not installed. Please install it to use this feature')
        extend_enum(cls, name, value)




def get_app_env(
    module_name: str,
) -> AppEnv:
    """
    Retrieves the app environment
    """
    module_name = module_name.replace(".", "_").upper()
    for key in {
        "SERVER_ENV",
        f"{module_name}_ENV",
        "APP_ENV",
        "ENVIRONMENT",
    }:
        if env_value := os.getenv(key):
            return AppEnv.from_env(env_value)

    from lzo.utils.system import is_in_kubernetes, get_host_name
    if is_in_kubernetes():
        # Name should be
        # scout-<service>-<index>
        # or 
        # scout-<service>-<env>-<index>
        hn = get_host_name()
        try:
            parts = hn.split("-")
            return AppEnv.from_env(parts[1]) if len(parts) > 2 else AppEnv.PRODUCTION
        except Exception as e:
            return AppEnv.from_hostname(hn)
        # parts = get_host_name().split("-")
        # return AppEnv.from_env(parts[2]) if len(parts) > 3 else AppEnv.PRODUCTION
    
    return AppEnv.LOCAL



from __future__ import annotations

"""
This module should generally not be used directly. 
It's only used when the module is known to be installed to provide
a consistent interface for the user.
"""

from lzl import load
load.LazyLoad("niquests", install_missing = True).__load__()
from niquests.exceptions import (
    ConnectionError,
    ConnectTimeout,
    FileModeWarning,
    HTTPError,
    JSONDecodeError,
    ReadTimeout,
    RequestException,
    RequestsDependencyWarning,
    Timeout,
    TooManyRedirects,
    URLRequired,
)


__all__ = [
    "ConnectionError",
    "ConnectTimeout",
    "FileModeWarning",
    "HTTPError",
    "JSONDecodeError",
    "ReadTimeout",
    "RequestException",
    "RequestsDependencyWarning",
    "Timeout",
    "TooManyRedirects",
    "URLRequired",
]


from __future__ import annotations

"""Connection presets that provide ergonomic defaults for :class:`Client`."""

import os
import typing as t

import httpx

from .types import typed as ht

__all__ = ["PresetConfig", "Preset", "PresetMap", "get_preset"]

PresetConfig = t.Literal[
    'default', 
    'low', 
    'mid', 
    'high',
    'polling',
    'streaming',
    'scraping',
    'downloads',
]

_ncpu_cores = os.cpu_count()
_conn_unit = max(2, _ncpu_cores // 10)

class Preset(t.TypedDict):
    """Configuration blob describing httpx limits, timeouts, and retries."""
    name: str
    limits: ht.Limits
    timeout: ht.TimeoutTypes
    retries: t.Optional[int] = 2
    kwargs: t.Optional[t.Dict[str, t.Any]] = None

DefaultPreset = Preset(
    name = 'default',
    limits = ht.DEFAULT_LIMITS,
    timeout = ht.DEFAULT_TIMEOUT_CONFIG,
    retries = 2,
    kwargs = None,
)

LowPreset = Preset(
    name = 'low',
    limits = ht.Limits(
        max_connections = _conn_unit,
        max_keepalive_connections = _conn_unit,
        keepalive_expiry = _conn_unit * 2,
    ),
    timeout = ht.DEFAULT_TIMEOUT_CONFIG,
    retries = 2,
    kwargs = None,
)

MidPreset = Preset(
    name = 'mid',
    limits = ht.Limits(
        max_connections = _conn_unit * 2,
        max_keepalive_connections = _conn_unit * 2,
        keepalive_expiry = _conn_unit * 4,
    ),
    timeout = httpx.Timeout(timeout = 15.0),
    retries = 5,
    kwargs = {'follow_redirects': True},
)

HighPreset = Preset(
    name = 'high',
    limits = ht.Limits(
        max_connections = _conn_unit * 4,
        max_keepalive_connections = _conn_unit * 4,
        keepalive_expiry = _conn_unit * 8,
    ),
    timeout = httpx.Timeout(timeout = 60.0),
    retries = 10,
    kwargs = {'follow_redirects': True, 'disable_httpx_logger': True},
)

PollingPreset = Preset(
    name = 'polling',
    limits = ht.Limits(
        max_connections = _conn_unit * 4,
        max_keepalive_connections = _conn_unit * 4,
        keepalive_expiry = _conn_unit * 8,
    ),
    timeout = httpx.Timeout(timeout = 120.0),
    retries = 5,
    kwargs = {'follow_redirects': True, 'disable_httpx_logger': True},
)

StreamingPreset = Preset(
    name = 'streaming',
    limits = ht.Limits(
        max_connections = _conn_unit * 4,
        max_keepalive_connections = _conn_unit * 4,
        keepalive_expiry = _conn_unit * 8,
    ),
    timeout = httpx.Timeout(timeout = 240.0),
    retries = 5,
    kwargs = {'follow_redirects': True, 'disable_httpx_logger': True},
)

ScrapingPreset = Preset(
    name = 'scraping',
    limits = ht.Limits(
        max_connections = _conn_unit * 4,
        max_keepalive_connections = _conn_unit * 4,
        keepalive_expiry = _conn_unit * 8,
    ),
    timeout = httpx.Timeout(timeout = 300.0),
    retries = 10,
    kwargs = {'follow_redirects': True, 'verify': False, 'disable_httpx_logger': True},
)

DownloadsPreset = Preset(
    name = 'downloads',
    limits = ht.Limits(
        max_connections = _conn_unit * 4,
        max_keepalive_connections = _conn_unit * 4,
        keepalive_expiry = _conn_unit * 8,
    ),
    timeout = httpx.Timeout(timeout = 600.0),
    retries = 5,
    kwargs = {'follow_redirects': True, 'disable_httpx_logger': True},
)

PresetMap: t.Dict[str, Preset] = {
    'default': DefaultPreset,
    'low': LowPreset,
    'mid': MidPreset,
    'high': HighPreset,
    'polling': PollingPreset,
    'streaming': StreamingPreset,
    'scraping': ScrapingPreset,
    'downloads': DownloadsPreset,
}

def get_preset(name: PresetConfig) -> t.Optional[Preset]:
    """Return the preset definition for *name* if one is registered."""

    return PresetMap.get(name)



from __future__ import annotations

"""Convenience façade for LazyOps thread/async pooling utilities."""

from .base import (
    ThreadPool,
    amap_iterable,
    ensure_coro,
    get_concurrency_limit,
    is_coro_func,
    set_concurrency_limit,
)
from .utils import is_in_async_loop

__all__: list[str] = [
    "ThreadPool",
    "amap_iterable",
    "ensure_coro",
    "get_concurrency_limit",
    "is_coro_func",
    "is_in_async_loop",
    "set_concurrency_limit",
]



from __future__ import annotations

"""
Extended Types
"""

import contextlib
from enum import Enum
from sqlalchemy import types
from sqlalchemy.schema import Sequence
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.sql import operators
from sqlalchemy.dialects.postgresql import JSONB, ARRAY, JSON
from sqlalchemy.dialects.postgresql.base import ischema_names

from lzl import load
import typing as t

# if load.t.TYPE_CHECKING:
#     import fileio
# else:
#     fileio = load.LazyLoad("fileio", install_missing=True, install_options={'package': 'file-io'})

# if t.TYPE_CHECKING:
#     from fileio import FileLike

from lzl.io import File, FileLike

class FileField(types.TypeDecorator):

    impl = types.String
    cache_ok = True
    mysql_default_length = 255

    def coerce_compared_value(self, op, value):
        """
        Coerce the value to be compared.
        """
        return types.String(length = self.mysql_default_length) if op in (operators.like_op, operators.not_like_op) else self
    
    def process_bind_param(self, value: t.Optional[t.Union[str, 'FileLike']], dialect):
        """
        Return the value to be stored in the database.
        """
        if value is None: return None
        if isinstance(value, str): return value
        return value.as_posix() \
            if hasattr(value, 'as_posix') \
                else str(value)

    def process_result_value(self, value: t.Optional[t.Union[str, t.Any]], dialect):
        """
        Process the value retrieved from the database.
        """
        return File(value) if value is not None else None


class JsonString(types.TypeDecorator):
    """
    A String type that stores JSON data.
    """
    impl = types.String
    cache_ok = True
    hashable = True

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """
        Initializes the JsonString t.Type
        """
        super().__init__(*args, **kwargs)
        from lzl.io.ser import get_serializer
        self.ser = get_serializer('json')
        self.ser.enforce_string_value = True

    def load_dialect_impl(self, dialect):
        """
        Load the dialect implementation.
        """
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect) -> t.Optional[str]:
        """
        Process the value to be stored in the database.
        """
        if value is not None:
            value = self.ser.dumps(value)
        return value

    def process_result_value(self, value, dialect) ->  t.Optional[t.Union[t.Dict[str, t.Any], t.List[t.Dict[str, t.Any]]]]:
        """
        Process the value retrieved from the database.
        """
        if value is not None:
            value = self.ser.loads(value)
        return value


class SerializedBinary(types.TypeDecorator):
    """
    A LargeBinary type that stores serialized data.
    """
    impl = types.LargeBinary
    cache_ok = True
    hashable = True

    def __init__(
        self,
        *args,
        serializer: t.Optional[str] = 'json',
        compression: t.Optional[str] = 'zstd',
        compression_level: int = 3,
        validate_raw: t.Optional[bool] = True,
        **kwargs,
    ) -> None:
        """
        Initializes the SerializedBinary t.Type

        - Supports Serialization: json, pickle, msgpack
        - Supports Compression: zstd, lz4, zlib, and bz2
        """
        super().__init__(*args, **kwargs)
        from lzl.io.ser import get_serializer
        self.ser = get_serializer(serializer, compression = compression, compression_level = compression_level)
        self.validate_raw = validate_raw

    
    def load_dialect_impl(self, dialect):
        """
        Load the dialect implementation.
        """
        return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value: t.Union[str, bytes, t.Any], dialect) -> t.Optional[bytes]:
        """
        Process the value to be stored in the database.
        """
        if value is not None:
            if self.validate_raw:
                if isinstance(value, bytes): return value
                if isinstance(value, str): return value.encode()
            value = self.ser.dumps(value)
        return value

    def process_result_value(self, value, dialect) ->  t.Optional[t.Union[t.Dict[str, t.Any], t.List[t.Dict[str, t.Any]]]]:
        """
        Process the value retrieved from the database.
        """
        if value is not None: value = self.ser.loads(value)
        return value



DateTimeUTC: types.DateTime = types.DateTime(timezone = True)

IntList: ARRAY = ARRAY(types.Integer, dimensions = 1)
BigIntList: ARRAY = ARRAY(types.BigInteger, dimensions = 1)
StringList: ARRAY = ARRAY(types.Text, dimensions = 1)
DateTimeUTCList: ARRAY = ARRAY(DateTimeUTC, dimensions = 1)
# AutoSequence: types.BigInteger = 

with contextlib.suppress(ImportError):
    from sqlalchemy_json import NestedMutableJson, MutableDict, mutable_json_type

    JsonStringField: t.Type[MutableDict] = mutable_json_type(dbtype = JsonString, nested=True)
    JsonStringNestedField: t.Type[NestedMutableJson] = mutable_json_type(dbtype = JsonString, nested = True) # type: ignore


    JSONBField: t.Type[MutableDict] = mutable_json_type(dbtype = JSONB)
    JSONBNestedField: t.Type[NestedMutableJson] = mutable_json_type(dbtype = JSONB, nested = True) # type: ignore


# for reflection
ischema_names['file'] = FileField
ischema_names['jsonstring'] = JsonString
ischema_names['serializedbinary'] = SerializedBinary

# Try to import pgvector
try:
    from pgvector.sqlalchemy import Vector

except ImportError:
    import numpy as np

    def from_db(value):
        # could be ndarray if already cast by lower-level driver
        if value is None or isinstance(value, np.ndarray):
            return value
        return np.array(value[1:-1].split(','), dtype=np.float32)
        
    def to_db(value, dim=None):
        if value is None: return value
        if isinstance(value, np.ndarray):
            if value.ndim != 1:
                raise ValueError('expected ndim to be 1')
            if not np.issubdtype(value.dtype, np.integer) and not np.issubdtype(value.dtype, np.floating):
                raise ValueError('dtype must be numeric')
            value = value.tolist()
        if dim is not None and len(value) != dim:
            raise ValueError('expected %d dimensions, not %d' % (dim, len(value)))
        return '[' + ','.join([str(float(v)) for v in value]) + ']'


    class Vector(types.UserDefinedType):
        cache_ok = True
        _string = types.String()

        def __init__(self, dim=None):
            super(types.UserDefinedType, self).__init__()
            self.dim = dim

        def get_col_spec(self, **kw):
            return "VECTOR" if self.dim is None else "VECTOR(%d)" % self.dim

        def bind_processor(self, dialect):
            def process(value):
                return to_db(value, self.dim)
            return process

        def literal_processor(self, dialect):
            string_literal_processor = self._string._cached_literal_processor(dialect)
            def process(value):
                return string_literal_processor(to_db(value, self.dim))
            return process

        def result_processor(self, dialect, coltype):
            def process(value):
                return from_db(value)
            return process

        class comparator_factory(types.UserDefinedType.Comparator):
            def l2_distance(self, other):
                return self.op('<->', return_type=types.Float)(other)

            def max_inner_product(self, other):
                return self.op('<#>', return_type=types.Float)(other)

            def cosine_distance(self, other):
                return self.op('<=>', return_type=types.Float)(other)

    ischema_names['vector'] = Vector

# from pgvector.asyncpg import Vector

# https://github.com/pgvector/pgvector


class VectorDistance(str, Enum):
    """
    Supported distance functions are:

    <-> - L2 distance
    <#> - (negative) inner product
    <=> - cosine distance
    <+> - L1 distance (added in 0.7.0)
    <~> - Hamming distance (binary vectors, added in 0.7.0)
    <%> - Jaccard distance (binary vectors, added in 0.7.0)
    """
    L2_DISTANCE = '<->'
    INNER_PRODUCT = '<#>'
    COSINE_DISTANCE = '<=>'
    L1_DISTANCE = '<+>'
    HAMMING_DISTANCE = '<~>'
    JACCARD_DISTANCE = '<%>'

    @property
    def vector_op(self) -> str:
        """
        Returns the Vector Op Name

        vector_l2_ops = L2 distance
        vector_ip_ops = (Negative) Inner Product
        vector_cosine_ops = Cosine Distance
        vector_l1_ops = L1 distance
        vector_hamming_ops = Hamming distance
        vector_jaccard_ops = Jaccard distance
        """
        if self == VectorDistance.L2_DISTANCE:
            return 'vector_l2_ops'
        elif self == VectorDistance.INNER_PRODUCT:
            return 'vector_ip_ops'
        elif self == VectorDistance.COSINE_DISTANCE:
            return 'vector_cosine_ops'
        elif self == VectorDistance.L1_DISTANCE:
            return 'vector_l1_ops'
        elif self == VectorDistance.HAMMING_DISTANCE:
            return 'vector_hamming_ops'
        elif self == VectorDistance.JACCARD_DISTANCE:
            return 'vector_jaccard_ops'
        return 'vector_ip_ops'

    
    @classmethod
    def parse(cls, value: str) -> 'VectorDistance':
        """
        Parses the Vector Distance
        """
        value = value.lower()
        if value in {'euclidean', 'l2_distance', 'l2', 'vector_l2_ops', '<->'}:
            return cls.L2_DISTANCE
        if value in {'inner_product', 'ip', 'negative_inner_product', 'max_inner_product', 'vector_ip_ops', '<#>'}:
            return cls.INNER_PRODUCT
        if value in {'cosine_distance', 'cosine', 'cosine_similarity', 'vector_cosine_ops', '<=>'}:
            return cls.COSINE_DISTANCE
        if value in {'l1', 'vector_l1_ops', '<+>'}:
            return cls.L1_DISTANCE
        if value in {'hamming', 'vector_hamming_ops', '<~>'}:
            return cls.HAMMING_DISTANCE
        if value in {'jaccard', 'vector_jaccard_ops', '<%>'}:
            return cls.JACCARD_DISTANCE
        raise ValueError(f"Invalid/Unsupported Vector Distance: {value}")



from __future__ import annotations

"""
The Base Config for the Oauth2 Provider
"""

import base64
import hashlib
from lzl.load import lazy_import
from lzl.logging import logger
from lzl.types import BaseSettings, eproperty
from pydantic import PrivateAttr, model_validator, field_validator, Field
from urllib.parse import urlparse, urlunparse, urljoin, parse_qsl, urlencode, quote
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated, Generator
from .token import AccessToken, IDToken, OAuth2Token


class BaseOAuth2Config(BaseSettings):
    """
    The Base Config for the Oauth2 Provider
    """

    name: Optional[str] = None
    issuer_url: Optional[str] = None
    jwt_url: Optional[str] = '/.well-known/jwks.json'
    jwt_issuer_url: Optional[str] = None

    client_id: Optional[str] = None
    client_secret: Optional[str] = None

    m2m_client_id: Optional[str] = None
    m2m_client_secret: Optional[str] = None

    code_verifier: Optional[str] = None
    code_challenge_method: Optional[str] = 'S256'
    scopes: Optional[List[str]] = None

    cache_base_key: Optional[str] = None
    cache_expiration: Optional[int] = 60 * 60 * 24 * 7 * 4 * 3 # 3 months
    cache_config: Optional[Dict[str, Any]] = Field(default_factory = dict)

    authorize_disabled: Optional[bool] = None
    user_info_retrieve_disabled: Optional[bool] = None
    
    if TYPE_CHECKING:
        access_token_class: Optional[Type[AccessToken]]
        id_token_class: Optional[Type[IDToken]]
        oauth2_token_class: Optional[Type[OAuth2Token]]
    else:
        access_token_class: Optional[Union[str, Any]] = None
        id_token_class: Optional[Union[str, Any]] = None
        oauth2_token_class: Optional[Union[str, Any]] = None

    @eproperty
    def m2m_enabled(self) -> bool:
        """
        Returns whether m2m is enabled
        """
        return bool(self.m2m_client_id and self.m2m_client_secret)
    
    @eproperty
    def m2m_audience(self) -> str:
        """
        Returns the m2m audience
        """
        return self.issuer_url

    @eproperty
    def scope(self) -> Optional[str]:
        """
        Returns the scope
        """
        return ' '.join(self.scopes) if self.scopes else None

    @eproperty
    def authorization_endpoint(self) -> str:
        """
        Returns the authorization endpoint
        """
        return urljoin(self.issuer_url, '/oauth2/auth')
    

    @eproperty
    def token_endpoint(self) -> str:
        """
        Returns the token endpoint
        """
        return urljoin(self.issuer_url, '/oauth2/token')
    

    @eproperty
    def logout_endpoint(self) -> str:
        """
        Returns the logout endpoint
        """
        return urljoin(self.issuer_url, '/logout')
    

    @classmethod
    def get_default_token_cls(cls, name: str) -> Union[Type[AccessToken], Type[IDToken], Type[OAuth2Token]]:
        """
        Returns the default token class
        """
        if name == 'access_token':
            return AccessToken
        return IDToken if name == 'id_token' else OAuth2Token

    @model_validator(mode = 'after')
    def validate_provider_config(self):
        """
        Validates the provider config
        """
        if self.jwt_url and 'https://' not in self.jwt_url:
            self.jwt_url = urljoin(self.issuer_url, self.jwt_url)
        if not self.jwt_issuer_url and self.issuer_url:
            self.jwt_issuer_url = self.issuer_url

        # Validate the token classes
        if self.access_token_class is None:
            self.access_token_class = self.get_default_token_cls('access_token')
        elif isinstance(self.access_token_class, str):
            self.access_token_class = lazy_import(self.access_token_class)
        
        if self.id_token_class is None:
            self.id_token_class = self.get_default_token_cls('id_token')
        elif isinstance(self.id_token_class, str):
            self.id_token_class = lazy_import(self.id_token_class)
        
        if self.oauth2_token_class is None:
            self.oauth2_token_class = self.get_default_token_cls('oauth2_token')
        elif isinstance(self.oauth2_token_class, str):
            self.oauth2_token_class = lazy_import(self.oauth2_token_class)
        return self


    def get_scopes(
        self,
        scopes: Optional[List[str]] = None,
        sep: Optional[str] = ' ',
    ) -> str:
        """
        Returns the scopes
        """
        return sep.join(scopes) if scopes else self.scope


    @staticmethod
    def create_code_challenge(secret_key: str) -> str:
        """
        Creates a code challenge
        """
        return base64.urlsafe_b64encode(hashlib.sha256(secret_key.encode()).digest()).decode().rstrip('=')

    @staticmethod
    def to_native(x: Union[str, bytes], encoding='ascii'):
        return x if isinstance(x, str) else x.decode(encoding)
    
    @staticmethod
    def to_bytes(x: Union[str, bytes, int, float], charset='utf-8', errors='strict'):
        if x is None: return None
        if isinstance(x, bytes): return x
        if isinstance(x, str): return x.encode(charset, errors)
        if isinstance(x, (int, float)): return str(x).encode(charset, errors)
        return bytes(x)

    @classmethod
    def encode_client_secret_basic(cls, client_id: str, client_secret: str) -> str:
        """
        Encodes the client secret basic
        """
        text = f'{quote(client_id)}:{quote(client_secret)}'
        return cls.to_native(base64.b64encode(cls.to_bytes(text, 'latin1')))


    def get_headers(
        self,
        endpoint_type: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the headers
        """
        return {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Accept': 'application/json',
        }
    

    def get_auth_params(
        self,
        redirect_url: str,
        scope: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the auth params
        """
        scope = scope or self.get_scopes(self.scopes)
        return {
            'redirect_uri': redirect_url,
            'client_id': self.client_id,
            'scope': scope,
            'access_type': 'offline',
            'prompt': 'consent',
            'response_type': 'code',
            **kwargs,
        }
    
    def get_fetch_token_params(
        self,
        code: str,
        redirect_url: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the fetch token params
        """
        return {
            'code': code,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'redirect_uri': redirect_url,
            'grant_type': 'authorization_code',
        }
    
    def get_refresh_token_params(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the refresh token params
        """
        return {
            'refresh_token': refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
        }
    


    def get_m2m_token_params(
        self,
        audience: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the m2m token params
        """
        audience = audience or self.m2m_audience
        return {
            'grant_type': 'client_credentials',
            'client_id': self.m2m_client_id,
            'client_secret': self.m2m_client_secret,
            'audience': audience,
        }
    


from __future__ import annotations

"""Public façade for LazyOps logging helpers.

The logging package wraps Loguru with project-specific defaults, colour
configuration, and filtering utilities.  This module re-exports the key entry
points so downstream consumers can import everything from a single location –
particularly useful when generating documentation via Mintlify.
"""

from .static import (
    DEFAULT_STATUS_COLORS,
    QUEUE_STATUS_COLORS,
    STATUS_COLOR,
    FALLBACK_STATUS_COLOR,
    DEFAULT_FUNCTION_COLOR,
    DEFAULT_CLASS_COLOR,
    RESET_COLOR,
    LOGLEVEL_MAPPING,
    REVERSE_LOGLEVEL_MAPPING,
    COLORED_MESSAGE_MAP,
)
from .state import (
    register_logger_module,
    is_registered_logger_module,
    add_api_log_filters,
    set_global_muted,
)
from .main import (
    NullLogger,
    Logger,
    create_default_logger,
    change_logger_level,
    get_logger,
    default_logger,
    null_logger,
    logger,
)

__all__ = [
    "DEFAULT_STATUS_COLORS",
    "QUEUE_STATUS_COLORS",
    "STATUS_COLOR",
    "FALLBACK_STATUS_COLOR",
    "DEFAULT_FUNCTION_COLOR",
    "DEFAULT_CLASS_COLOR",
    "RESET_COLOR",
    "LOGLEVEL_MAPPING",
    "REVERSE_LOGLEVEL_MAPPING",
    "COLORED_MESSAGE_MAP",
    "register_logger_module",
    "is_registered_logger_module",
    "add_api_log_filters",
    "set_global_muted",
    "NullLogger",
    "Logger",
    "create_default_logger",
    "change_logger_level",
    "get_logger",
    "default_logger",
    "null_logger",
    "logger",
]





from .logs import logger
from .helpers import (
    is_naive,
    total_seconds,
    remove_trailing_slash,
    parse_stream,
    aparse_stream,
    weighted_choice,
    
)

from .tokenization import (
    modelname_to_contextsize,
    get_token_count,
    get_max_tokens,
    get_chat_tokens_count,
    get_max_chat_tokens,
    fast_tokenize,
)

from .resolvers import fix_json
from .fixjson import resolve_json



from __future__ import annotations

"""
The Hatchet Context Object

Patched: 2024-08-08 19:02:25
Version: 0.32.0
Last Version: 0.31.0

"""

import copy
from concurrent.futures import ThreadPoolExecutor
from hatchet_sdk.context.context import Context as BaseContext, ContextAioImpl
# from hatchet_sdk.context import Context as BaseContext
from hatchet_sdk.logger import logger
from lazyops.libs.abcs.utils.format_utils import clean_html
from ..utils import json_serializer
from typing import TypeVar, Dict, Any, Optional, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from hatchet_sdk.clients.dispatcher import Action
    from hatchet_sdk.clients.events import EventClientImpl
    from hatchet_sdk.clients.run_event_listener import RunEventListenerClient
    from hatchet_sdk.clients.workflow_listener import PooledWorkflowRunListener
    from hatchet_sdk.clients.admin import AdminClientImpl
    from hatchet_sdk.clients.dispatcher import Action, DispatcherClientImpl
    from hatchet_sdk.context.worker_context import WorkerContext

ContextT = TypeVar('ContextT', bound = 'Context')

class Context(BaseContext):
    """
    The Context Object
    """

    if TYPE_CHECKING:
        input: Dict[str, Any]

    def __init__(
        self,
        action: 'Action',
        dispatcher_client: 'DispatcherClientImpl',
        admin_client: 'AdminClientImpl',
        event_client: 'EventClientImpl',
        workflow_listener: 'PooledWorkflowRunListener',
        workflow_run_event_listener: 'RunEventListenerClient',
        worker: 'WorkerContext',
        namespace: str = "",
    ):
        self.worker = worker

        self.aio = ContextAioImpl(
            action,
            dispatcher_client,
            admin_client,
            event_client,
            workflow_listener,
            workflow_run_event_listener,
            worker,
            namespace,
        )

        # Check the type of action.action_payload before attempting to load it as JSON
        if isinstance(action.action_payload, (str, bytes, bytearray)):
            try:
                self.data = json_serializer.loads(action.action_payload)
            except Exception as e:
                logger.error(f"Error parsing action payload: {e}")
                # Assign an empty dictionary if parsing fails
                self.data = {}
        else:
            # Directly assign the payload to self.data if it's already a dict
            self.data = (
                action.action_payload if isinstance(action.action_payload, dict) else {}
            )

        self.action = action
        # FIXME: stepRunId is a legacy field, we should remove it
        self.stepRunId = action.step_run_id

        self.step_run_id = action.step_run_id
        self.exit_flag = False
        self.dispatcher_client = dispatcher_client
        self.admin_client = admin_client
        self.event_client = event_client
        self.workflow_listener = workflow_listener
        self.workflow_run_event_listener = workflow_run_event_listener
        self.namespace = namespace

        # FIXME: this limits the number of concurrent log requests to 1, which means we can do about
        # 100 log lines per second but this depends on network.
        self.logger_thread_pool = ThreadPoolExecutor(max_workers=1)
        self.stream_event_thread_pool = ThreadPoolExecutor(max_workers=1)

        # store each key in the overrides field in a lookup table
        # overrides_data is a dictionary of key-value pairs
        self.overrides_data = self.data.get("overrides", {})

        if action.get_group_key_run_id != "":
            self.input = self.data
        else:
            self.input = self.data.get("input", {})


    @property
    def parents(self) -> Dict[str, Dict[str, Union[str, List[Any], Dict[str, Union[str, Dict, List]]]]]:
        """
        Gets the parents
        """
        return self.data.get('parents', {})

    def get_kwargs(self) -> Dict[str, Any]:
        """
        Gets the kwargs
        """
        return copy.deepcopy(self.input)

    @property
    def event_trigger(self) -> str:
        """
        Gets the event trigger
        """
        return self.input.get('trigger', self.data.get('triggered_by', 'manual'))

    
    @property
    def is_cron(self) -> bool:
        """
        Checks if the event is a cron event
        """
        return self.event_trigger == 'cron'
    
    @property
    def workflow_id(self) -> str:
        """
        Gets the workflow run id
        """
        return self.action.workflow_run_id
    
    def log(self, line: str):
        """
        Logs a line
        """
        # Need to clean up any colored formatting
        # Remove the last 4 characters
        line = clean_html(line)
        if line.endswith('[0m'): line = line[:-4]
        return super().log(line)
    
    def get(self, key: str, default: Any = None) -> Any:
        """
        Gets the value
        """
        return self.input.get(key, default)


from __future__ import annotations

"""
Base Helpers
"""

import asyncio
import collections.abc
import contextlib
import functools
import hashlib
import importlib.util
import inspect
import json
import pathlib
import random
import string
import time
import typing as t

def extract_function_kwargs(
    func: t.Callable[..., t.Any],
    **kwargs: t.Any,
) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
    """Split ``kwargs`` into arguments accepted by ``func`` and leftovers."""

    func_kwargs = {k: v for k, v in kwargs.items() if k in func.__code__.co_varnames}
    extra_kwargs = {k: v for k, v in kwargs.items() if k not in func_kwargs}
    return func_kwargs, extra_kwargs

    # return {k: v for k, v in kwargs.items() if k not in func.__code__.co_varnames}


def is_coro_func(obj: t.Any, func_name: t.Optional[str] = None) -> bool:
    """Return ``True`` when ``obj`` or ``obj.func_name`` is awaitable."""
    try:
        if inspect.iscoroutinefunction(obj): return True
        if inspect.isawaitable(obj): return True
        if func_name and hasattr(obj, func_name) and inspect.iscoroutinefunction(getattr(obj, func_name)):
            return True
        return bool(hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__))

    except Exception:
        return False


def exponential_backoff(
    attempts: int,
    base_delay: int = 1,
    max_delay: t.Optional[int] = None,
    jitter: bool = True,
) -> float:
    """Compute an exponentially increasing delay for retry loops."""
    if max_delay is None:
        max_delay = float("inf")
    backoff = min(max_delay, base_delay * 2 ** max(attempts - 1, 0))
    if jitter:
        backoff = backoff * random.random()
    return backoff

def retryable(limit: int = 3, delay: int = 3) -> t.Callable[[t.Callable[..., t.Any]], t.Callable[..., t.Any]]:
    """Wrap a callable to retry transient failures with exponential backoff."""

    def decorator(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        if not inspect.iscoroutinefunction(func):

            def sync_wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
                for n in range(limit - 1):
                    with contextlib.suppress(Exception):
                        return func(*args, **kwargs)
                    time.sleep(exponential_backoff(n, base_delay=delay))
                return func(*args, **kwargs)

            return sync_wrapper

        async def async_wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            for n in range(limit - 1):
                with contextlib.suppress(Exception):
                    return await func(*args, **kwargs)
                await asyncio.sleep(exponential_backoff(n, base_delay=delay))
            return await func(*args, **kwargs)

        return async_wrapper

    return decorator


def suppress(*exceptions: t.Optional[t.Union[t.Tuple[Exception, ...], Exception]]) -> t.Callable[[t.Callable[..., t.Any]], t.Callable[..., t.Any]]:
    """Decorator factory mirroring :func:`contextlib.suppress` for callables."""

    if not exceptions:
        exceptions = (Exception,)

    def wrapper(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        if is_coro_func(func):

            @functools.wraps(func)
            async def wrapped_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
                with contextlib.suppress(*exceptions):
                    return await func(*args, **kwargs)

            return wrapped_func

        @functools.wraps(func)
        def wrapped_func(*args: t.Any, **kwargs: t.Any) -> t.Any:
            with contextlib.suppress(*exceptions):
                return func(*args, **kwargs)

        return wrapped_func

    return wrapper

def create_secret(nbytes: t.Optional[int] = 16) -> str:
    """Generate a hexadecimal token using ``secrets.token_hex``."""

    import secrets

    return secrets.token_hex(nbytes)


def create_unique_id(
    method: t.Optional[str] = 'uuid4',
    alph_only: bool = False,
    length: t.Optional[int] = None,
) -> str:
    """Generate a UUID-based identifier with optional filtering/truncation."""

    import uuid

    meth = getattr(uuid, method, None)
    if not meth:
        raise ValueError(f'Invalid UUID method: {method}')
    val = str(meth())
    if alph_only:
        val = ''.join(c for c in val if c.isalpha())
    if length:
        while len(val) < length:
            val += str(meth())
            if alph_only:
                val = ''.join(c for c in val if c.isalpha())
            if val.endswith('-'):
                val = val[:-1]
        val = val[:length]
    return val

ALPHA = string.ascii_letters

def create_unique_secret_key(length: int = 44) -> str:
    """Return an uppercase alphanumeric key suitable for shared secrets."""

    return ''.join(random.choice(ALPHA) for _ in range(length)).upper()


def fail_after(delay: t.Union[int, float] = 5.0) -> t.Callable[[t.Callable[..., t.Any]], t.Callable[..., t.Any]]:
    """Raise a timeout if ``func`` does not finish within ``delay`` seconds."""

    import concurrent.futures

    def decorator(func: t.Callable[..., t.Any]) -> t.Callable[..., t.Any]:
        @functools.wraps(func)
        def wrapper(*args: t.Any, **kwargs: t.Any) -> t.Any:
            executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
            try:
                future = executor.submit(func, *args, **kwargs)
                return future.result(timeout=delay)
            finally:
                executor.shutdown(wait=False, cancel_futures=True)

        return wrapper

    return decorator

    

_PRIMITIVE_TYPES = (str, int, float, bool, bytes)


def _is_primitive(value: t.Any) -> bool:
    return value is None or isinstance(value, _PRIMITIVE_TYPES)


def _normalized_value(value: t.Any) -> t.Any:
    if _is_primitive(value):
        return value
    if isinstance(value, collections.abc.Mapping):
        return tuple(sorted((k, _normalized_value(v)) for k, v in value.items()))
    if isinstance(value, (list, tuple)):
        return tuple(_normalized_value(item) for item in value)
    if isinstance(value, set):
        return tuple(sorted(_normalized_value(item) for item in value))
    if isinstance(value, collections.abc.Iterable) and not isinstance(value, (bytes, bytearray)):
        return tuple(_normalized_value(item) for item in value)
    return repr(value)


def _hash_value(value: t.Any) -> str:
    normalized = _normalized_value(value)
    if isinstance(normalized, t.Hashable):
        return str(hash(normalized))
    serialized = json.dumps(normalized, sort_keys=True, default=repr)
    return hashlib.sha256(serialized.encode('utf-8')).hexdigest()


def _deduplicate_list(items: t.Iterable[t.Any]) -> t.List[t.Any]:
    seen: t.Set[t.Hashable] = set()
    unique: t.List[t.Any] = []
    for item in items:
        if _is_primitive(item):
            key = t.cast(t.Hashable, item)
        else:
            key = _hash_value(item)
        if key in seen:
            continue
        seen.add(key)
        unique.append(item)
    return unique


def update_dict(
    d: t.Dict[t.Hashable, t.Any],
    u: t.Mapping[t.Hashable, t.Any],
    exclude_none: bool = False,
    unset_value: t.Optional[str] = 'UNSET',
    deduplicate: bool = True,
) -> t.Dict[t.Hashable, t.Any]:
    """Recursively update a dictionary, respecting unset sentinels."""
    unset_keys = []
    for k, v in u.items():
        if exclude_none and v is None:
            continue
        if isinstance(v, str) and v == unset_value:
            unset_keys.append(k)
            continue
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_dict(
                d.get(k, {}),
                v,
                exclude_none=exclude_none,
                unset_value=unset_value,
                deduplicate=deduplicate,
            )
        elif isinstance(v, list):
            existing = d.get(k, [])
            if not isinstance(existing, list):
                existing = []
            combined = existing + v
            if deduplicate:
                combined = _deduplicate_list(combined)
            d[k] = combined
        else:
            d[k] = v
    for k in unset_keys:
        d.pop(k, None)
    return d


def merge_recursive_dict(
    d: t.Mapping[str, t.Any],
    current_key: str,
) -> t.Dict[str, t.Any]:
    """Flatten nested dictionaries into dotted-key mappings."""
    if not isinstance(d, collections.abc.Mapping):
        return d
    mapping = {}
    for k, v in d.items():
        current_key = f'{current_key}.{k}'
        if isinstance(v, (collections.abc.Mapping, dict)):
            mapping[current_key] = merge_recursive_dict(v, current_key)
        else:
            mapping[current_key] = v
    return mapping


def flatten_dict_value(
    d: t.Mapping[str, t.Any],
    parent_key: str = '',
) -> t.Dict[str, t.Any]:
    """Return a flattened dictionary where keys are joined by ``.``."""

    items: t.List[t.Tuple[str, t.Any]] = []
    for k, v in d.items():
        new_key = f'{parent_key}.{k}' if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict_value(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)



# https://stackoverflow.com/questions/27265939/comparing-python-dictionaries-and-nested-dictionaries



def parse_deltas(deltas: t.Mapping[str, t.Tuple[t.Any, t.Any]]) -> t.Dict[str, t.Any]:
    """Normalise nested diff output into a flat mapping."""

    res: t.Dict[str, t.Any] = {}
    for k, v in deltas.items():
        if isinstance(v[0], dict):
            tmp = diff_dict(v[0], v[1])
            if tmp:
                res[k] = tmp
        else:
            res[k] = v[1]
    return res

def diff_dict(d1: t.Mapping[str, t.Any], d2: t.Mapping[str, t.Any]) -> t.Dict[str, t.Any]:
    """
    Returns the difference between two dictionaries

    d1 should be the original dict
    d2 should be the new/updated dict

    Returns a dict with the differences between d1 and d2
    """
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    shared_keys = d1_keys.intersection(d2_keys)
    shared_deltas = {o: (d1[o], d2[o]) for o in shared_keys if d1[o] != d2[o]}
    added_keys = d2_keys - d1_keys
    added_deltas = {o: (None, d2[o]) for o in added_keys}
    deltas = {**shared_deltas, **added_deltas}
    return parse_deltas(deltas)



@functools.lru_cache()
def get_module_path(
    module_name: str, 
    **kwargs
) -> pathlib.Path:
    """Locate the filesystem path for ``module_name`` using import metadata."""

    module_spec = importlib.util.find_spec(module_name)
    if not module_spec:
        raise ValueError(f"Module {module_name} not found")

    for path in module_spec.submodule_search_locations:
        module_path = pathlib.Path(path)
        if module_path.exists(): return module_path
    
    raise ValueError(f"Module {module_name} cant be found in the path")



from __future__ import annotations

"""
The Application State
"""
import os
import abc
import atexit
import contextlib
import pathlib
from lzl.logging import logger
from lzl.types import BaseModel, Field
from typing import Optional, Dict, Any, List, Union, Generator, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from lzo.types.base import RegisteredSettings as BaseSettings
    from lzl.io import TemporaryData as StateData


class AppState(BaseModel):
    """
    Holds the state of the current settings
    """
    configured: Set[str] = Field(default_factory=set)
    initialized: Set[str] = Field(default_factory=set)
    completed: bool = False

    app_entrypoint: Optional[str] = None
    worker_entrypoint: Optional[str] = None

    k8s_kubeconfigs: Optional[Dict[str, str]] = Field(default_factory=dict)
    k8s_active_ctx: Optional[str] = None

    app_module_name: Optional[str] = None
    data_path: Optional[pathlib.Path] = None
    
    ctx: Optional[Dict[str, Any]] = {}


    @property
    def is_silent(self) -> bool:
        """
        Returns whether the current state is silent
        """
        return self.ctx.get('silent', False)
    
    @property
    def settings(self) -> 'BaseSettings':
        """
        Returns the settings
        """
        return self.ctx.get('settings', None)

    @property
    def stx(self) -> Optional['StateData']:
        """
        Returns the StateData
        """
        return self.ctx.get('stx', None)

    @property
    def process_id(self) -> Optional[int]:
        """
        Returns the process id
        """
        return self.ctx.get('process_id', None)
    
    @property
    def is_primary_process(self) -> bool:
        """
        Returns whether this is the primary process
        """
        return self.process_id == self.stx.get('primary_process_id', 0)
    
    @property
    def is_primary_server_process(self) -> bool:
        """
        Returns whether this is the primary server process
        """
        return self.process_id == self.stx.get('primary_server_process_id', 0)

    @property
    def is_leader_process(self) -> bool:
        """
        Returns whether this is the leader process
        """
        return self.process_id in self.stx.get('leader_process_ids', []) or self.is_primary_process

    @property
    def server_process_id_path(self) -> pathlib.Path:
        """
        Returns the server process id path
        """
        return self.data_path.joinpath(f'{self.app_module_name}.pid')

    def configure_silent(self, _silent: Optional[bool] = None):
        """
        Configures the silent mode
        """
        if _silent is not None:
            self.ctx['silent'] = _silent

    def get_kubeconfig(
        self, 
        name: Optional[str] = None, 
        set_as_envval: Optional[bool] = True,
        set_active: Optional[bool] = False,
    ) -> str:
        """
        Returns the kubeconfig for the given context
        """
        name = name or self.k8s_active_ctx
        if name is not None and name in self.k8s_kubeconfigs:
            kconfig = self.k8s_kubeconfigs[name]
        else:
            from lzo.utils.system import get_local_kubeconfig
            kconfig = get_local_kubeconfig(name = name, set_as_envval = False)
            if not name: name = pathlib.Path(kconfig).stem
            self.k8s_kubeconfigs[name] = kconfig
        if set_as_envval: os.environ['KUBECONFIG'] = kconfig
        if not self.k8s_active_ctx or set_active: self.k8s_active_ctx = name
        return kconfig

    def bind_settings(self, settings: 'BaseSettings'):
        """
        Binds the settings to this state
        """
        # puts it in here so that we avoid type checking at the module level.
        self.ctx['settings'] = settings
        self.ctx['process_id'] = os.getpid()
        if hasattr(settings, 'app_module_name'):
            self.app_module_name = settings.app_module_name
        else:
            self.app_module_name = settings.__class__.__module__.split(".")[0]
        if hasattr(settings, 'data_path'):
            self.data_path = settings.data_path
        else:
            from lzo.utils.assets import get_module_path
            module_path = get_module_path(self.app_module_name)
            self.data_path = module_path.joinpath('.data')
        self.configure_stx()

    
    def on_exit(self):
        """
        Called on exit
        """
        if 'stx' in self.ctx and self.process_id == self.stx['primary_process_id']:
            logger.info(f"Removing AppState File: {self.ctx['stx_filepath']}", colored = True, prefix = f"|r|State: {self.app_module_name}|e|")
            with contextlib.suppress(FileNotFoundError):
                self.stx.close()
                os.unlink(self.ctx['stx_filepath'])
                os.unlink(self.stx.filelock_path.as_posix())

    def configure_stx(self):
        """
        Configures the stateful statefuldata
        """
        if 'stx' not in self.ctx:
            stx_filepath = self.data_path.joinpath(f'{self.app_module_name}.state.json')
            self.ctx['stx_filepath'] = stx_filepath.as_posix()
            from lzl.io.persistence import TemporaryData as StateData
            stx = StateData(filepath = stx_filepath)
            self.ctx['stx'] = stx
            atexit.register(self.on_exit)

    def set_primary_process_id(self, process_id: int):
        # sourcery skip: class-extract-method
        """
        Sets the primary process id
        """
        self.stx['primary_process_id'] = process_id
    
    def set_primary_server_process_id(self, process_id: Optional[int] = None):
        """
        Sets the primary server process id
        """
        # Try to find it
        if 'primary_server_process_id' in self.stx.keys(): return
        if process_id is None and self.server_process_id_path.exists():
            with contextlib.suppress(Exception):
                process_id = int(self.server_process_id_path.read_text())
        if process_id is None:
            return
        self.stx['primary_server_process_id'] = process_id
        logger.info(f"Primary Server Process ID: {process_id}", colored = True, prefix = "|g|State|e|")
    
    def add_leader_process_id(self, process_id: int, kind: Optional[str] = None):
        """
        Adds a leader process id
        """
        self.stx.append('leader_process_ids', process_id)

    def has_logged(self, key: str) -> bool:
        """
        Returns whether the key has been logged
        """
        return self.stx.append('logged', key)
        



from .config import PostgresConfig, PostgresSettings
from .backend import PostgresBackend, SQLAlchemyBackend, SQLModelBackend, PostgresBackendMap, PostgresBackendClasses
from .manager import PostgresBackendManager


from __future__ import annotations

"""
Global State Objects that can be used to share state across modules
"""

import pathlib
import typing as t
from lzl.proxied import ProxyObject

if t.TYPE_CHECKING:
    from lzl.io import TemporaryData

lib_path = pathlib.Path(__file__).parent

_temporary_data: t.Optional['TemporaryData'] = None

# This temp data will be at lzl/.data

def get_temp_data() -> 'TemporaryData':
    """
    Gets the temporary data object
    """
    global _temporary_data
    if _temporary_data is None:
        from lzl.io.persistence import TemporaryData
        _temporary_data = TemporaryData(lib_path.joinpath('.tmpdata'))
    return _temporary_data


TempData: 'TemporaryData' = ProxyObject(
    obj_getter = get_temp_data
)




from __future__ import annotations

"""
Base Hatchet Client with Modifications

Patched: {{ timestamp }}
Version: {{ version }}
Last Version: {{ last_version }}
"""

import os
import signal
import asyncio
import functools
import contextvars
import grpc
from threading import Thread
from concurrent.futures import Future, ThreadPoolExecutor
from hatchet_sdk.loader import ClientConfig
from hatchet_sdk.worker import (
    Worker as BaseWorker, 
    WorkerStatus, 
    errorWithTraceback, 
    copy_context_vars,
    wr, sr,
    new_dispatcher, 
    new_admin,
    new_listener,
    PooledWorkflowRunListener,
    GetActionListenerRequest,
    WorkerContext
)
from hatchet_sdk.client import new_client_raw
from threading import Thread, current_thread
from hatchet_sdk.clients.dispatcher import Action, ActionListenerImpl
from hatchet_sdk.dispatcher_pb2 import (
    GROUP_KEY_EVENT_TYPE_COMPLETED,
    GROUP_KEY_EVENT_TYPE_FAILED,
    GROUP_KEY_EVENT_TYPE_STARTED,
    STEP_EVENT_TYPE_COMPLETED,
    STEP_EVENT_TYPE_FAILED,
    STEP_EVENT_TYPE_STARTED,
    ActionType,
    GroupKeyActionEvent,
    GroupKeyActionEventType,
    StepActionEvent,
    StepActionEventType,
    WorkerUnsubscribeRequest,
)
from hatchet_sdk.logger import logger
from hatchet_sdk.metadata import get_metadata
from typing import Any, Dict, Callable, Type, Optional, TYPE_CHECKING
from .context import Context, ContextT
from .utils import new_session, json_serializer
from lzo.utils.logs import logger as _logger

if TYPE_CHECKING:
    from .session import HatchetSession

class Worker(BaseWorker):
    """
    The Worker Object
    """
    SIGNALS = [signal.SIGINT, signal.SIGTERM] if os.name != "nt" else [signal.SIGTERM]

    
{% for name, func in new_funcs.items() %}
{{ func | replace("json.dumps", "json_serializer.dumps") }}
{% endfor %}

    # @property
    # def loop(self) -> asyncio.AbstractEventLoop:
    #     """
    #     Returns the event loop
    #     """
    #     if not self._loop: 
    #         try:
    #             self._loop = asyncio.get_running_loop()
    #         except RuntimeError as e:
    #             logger.error(f"Error getting running loop: {e}")
    #             try:
    #                 self._loop = asyncio.get_event_loop()
    #             except RuntimeError as e:
    #                 logger.error(f"Error getting event loop: {e}")
    #                 self._loop = asyncio.new_event_loop()
    #                 asyncio.set_event_loop(self._loop)
    #     return self._loop
    
    # @loop.setter
    # def loop(self, value: asyncio.AbstractEventLoop):
    #     """
    #     Sets the loop
    #     """
    #     self._loop = value

    # def run(self):
    #     """
    #     Run function
    #     """
    #     self.main_task = self.loop.create_task(self.async_run)
    #     try:
    #         self.loop.run_until_complete(self.main_task)
    #     except asyncio.CancelledError:  # pragma: no cover
    #         # happens on shutdown, fine
    #         pass
    #     finally:
    #         self.loop.run_until_complete(self.exit_gracefully)

    async def watch_for_termination(self):
        """
        Watches for termination
        """
        self.event = asyncio.Event()
        
        try:
            for signum in self.SIGNALS: self.loop.add_signal_handler(signum, self.event.set)
            await self.event.wait()
        # except asyncio.CancelledError:  # pragma: no cover
            # happens on shutdown, fine
            # self.event.set()
        except Exception as e:
            _logger.trace('Error in Watching', e)
            # self.event.set()
        finally:
            self.event.set()
            logger.warning(f'Exiting!')
            self.loop.run_until_complete(self.exit_gracefully)

    def run_watcher_event(self):
        """
        Starts the watcher event
        """
        asyncio.create_task(self.watch_for_termination())
    
    async def async_run(self, **kwargs):
        """
        Async Function to run the worker
        """
        self.event = asyncio.Event()
        for signum in self.SIGNALS: self.loop.add_signal_handler(signum, self.event.set)
        task = None
        try:
            task = self.loop.create_task(self.async_start)
            await self.event.wait()
        except Exception as e:
            logger.error(f'Shutdown: {e}')
            raise e
        finally:
            self.event.set()
            if task is not None:
                if not task.done(): task.cancel()
            if self.main_task: self.main_task.cancel()
            await asyncio.wait_for(asyncio.gather(task), timeout = 5.0)








from __future__ import annotations

"""Helpers for lazily initialising higher-order function wrappers."""

import functools
import typing as t

ReturnT = t.TypeVar("ReturnT")


def lazy_function_wrapper(
    function: t.Callable[..., ReturnT],
    *function_args: t.Any,
    **function_kwargs: t.Any,
) -> t.Callable[[t.Callable[..., ReturnT]], t.Callable[..., ReturnT]]:
    """Defer wrapper creation until the wrapped function is first invoked.

    The initialiser ``function`` is executed at most once.  It may return a
    callable that itself accepts the original function and returns a wrapped
    version.  When ``function`` returns ``None`` the original callable is used
    as-is, preserving the legacy behaviour relied on by existing integrations.
    """

    _initialized = False
    _initialized_function: t.Optional[t.Callable[[t.Callable[..., ReturnT]], t.Callable[..., ReturnT]]] = None

    def lazywrapped(func: t.Callable[..., ReturnT]) -> t.Callable[..., ReturnT]:
        from lzl.pool import ThreadPool, is_coro_func

        if is_coro_func(func):
            @functools.wraps(func)
            async def _wrapper(*args: t.Any, **kwargs: t.Any) -> ReturnT:
                nonlocal _initialized, _initialized_function
                if not _initialized:
                    _initialized_function = await ThreadPool.asyncish(
                        function,
                        *function_args,
                        **function_kwargs,
                    )
                    _initialized = True
                if _initialized_function is None:
                    return await func(*args, **kwargs)
                return await _initialized_function(func)(*args, **kwargs)

            return _wrapper

        @functools.wraps(func)
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> ReturnT:
            nonlocal _initialized, _initialized_function
            if not _initialized:
                _initialized_function = function(*function_args, **function_kwargs)
                _initialized = True
            if _initialized_function is None:
                return func(*args, **kwargs)
            return _initialized_function(func)(*args, **kwargs)

        return _wrapper

    return lazywrapped



from __future__ import annotations

"""
Sqlite Manager
"""

import typing as t
from lzl.logging import logger
from ..base.manager import DBBackendManager
from .backend import BackendType, SqliteBackendClasses


class SqliteBackendManager(DBBackendManager):
    """
    Sqlite Backend Manager
    """

    db_type: t.Optional[str] = 'sqlite'

    if t.TYPE_CHECKING:
        import pathlib
        from .config import SqliteConfig, SqliteSettings

        def initialize_backend(
            self,
            name: t.Optional[str] = None,
            backend: t.Optional[BackendType] = None,

            settings: t.Optional[SqliteSettings] = None,
            config: t.Optional[SqliteConfig] = None,
            
            filepath: t.Optional[t.Union[str, pathlib.Path]] = None, 
            env_var: t.Optional[str] = None,

            settings_app_name: t.Optional[str] = None,
            settings_filepath: t.Optional[t.Union[str, pathlib.Path]] = None,
            settings_env_var: t.Optional[str] = None,

            overrides: t.Optional[t.Dict[str, t.Any]] = None,
            **kwargs,
        ) -> SqliteBackendClasses:
            """
            [Sqlite] Initializes the backend
            """
            ...


        def configure_backend(
            self,
            name: t.Optional[str] = None,
            backend: t.Optional[BackendType] = None,

            settings: t.Optional[SqliteSettings] = None,
            config: t.Optional[SqliteConfig] = None,
            
            filepath: t.Optional[t.Union[str, pathlib.Path]] = None, 
            env_var: t.Optional[str] = None,

            settings_app_name: t.Optional[str] = None,
            settings_filepath: t.Optional[t.Union[str, pathlib.Path]] = None,
            settings_env_var: t.Optional[str] = None,

            overrides: t.Optional[t.Dict[str, t.Any]] = None,
            set_as_current: t.Optional[bool] = None,
            **kwargs,
        ) -> SqliteBackendClasses:
            """
            [Sqlite] Configures the backend
            """
            ...

        @property
        def bkend(self) -> 'SqliteBackendClasses':
            """
            Returns the backend
            """
            ...


from __future__ import annotations

"""
TemporalIO Config: Worker
"""

import typing as t
from urllib.parse import urlparse
from lzo.types import BaseSettings, model_validator, Field
from lzl.types import eproperty
from lzl.proxied import ProxyObject
from ..utils import logger

if t.TYPE_CHECKING:
    from lzl.io.persistence import TemporaryData
    from temporalio.converter import DataConverter
    from ..registry import TemporalRegistry

class TemporalWorkerConfig(BaseSettings):
    """
    Temporal Worker Configuration
    """
    queue: str = Field(default="")
    name: str = Field(default="")
    identity: t.Optional[str] = Field(None, description = "Identity for this client. If unset, a default is created based on the version of the SDK.")
    namespace: t.Optional[str] = Field(default="")
    interceptors: t.Optional[t.List[str]] = Field(default=None)
    activities: t.Optional[t.List[str]] = Field(default_factory=list)
    workflows: t.Optional[t.List[str]] = Field(default_factory=list)
    
    converter: t.Optional[str] = Field(default=None)
    factory: t.Optional[str] = Field(default=None)
    pre_init: t.Optional[t.List[str]] = Field(default=None)
    max_concurrent_activities: int = Field(default=100)
    max_concurrent_workflow_tasks: int = Field(default=100)
    debug_mode: bool = Field(default=False)
    disable_eager_activity_execution: bool = Field(default=True) # pylint: disable=invalid-name
    metric_bind_address: str = Field(default="0.0.0.0:9000")

    mapping: t.Optional[t.Dict[str, str]] = Field(default=None, description = "Mapping of workflow names to activity names")

    class Config(BaseSettings.Config):
        env_prefix = "TEMPORAL_WORKER_"

    @model_validator(mode = 'before')
    def validate_config_values(cls, values: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        """
        Validates the values
        """
        if values.get('mapping') and 'workflows' not in values and 'activities' not in values:
            values['workflows'] = list(values['mapping'].keys())
            values['activities'] = list(values['mapping'].values())
        return values

            



    





from __future__ import annotations

"""
Hatchet Utilities
"""

import os
import pathlib
from hatchet_sdk.client import ClientImpl
from hatchet_sdk.loader import ClientConfig, ConfigLoader
from lzl.logging import logger
from lzo.utils.system import get_ulimits, set_ulimits
from kvdb.io.serializers import get_serializer
from typing import Any, Dict, Union, Optional, TYPE_CHECKING
from .version import CURRENT_VERSION as VERSION

if TYPE_CHECKING:
    from lzl.api.hatchet.config import HatchetSettings
    from lzl.api.hatchet.client import HatchetClient
    from lzl.api.hatchet.session import HatchetSession

    
lib_path = pathlib.Path(__file__).parent
json_serializer = get_serializer(
    'json',
    compression = None,
)

_hatchet_settings: Optional['HatchetSettings'] = None
_hatchet_client: Optional['HatchetClient'] = None
_hatchet_sessions: Dict[str, 'HatchetSession'] = {}

def get_hatchet_settings() -> 'HatchetSettings':
    """
    Retrieves the Hatchet settings
    """
    global _hatchet_settings
    if _hatchet_settings is None:
        from lzl.api.hatchet.config import HatchetSettings
        _hatchet_settings = HatchetSettings()
    return _hatchet_settings

def set_hatchet_settings(settings: 'HatchetSettings', override: bool = False):
    """
    Sets the global Hatchet Settings
    """
    global _hatchet_settings
    if _hatchet_settings is None or override:
        _hatchet_settings = settings

def get_hatchet_session(
    instance: str = 'default'
) -> Optional['HatchetSession']:
    """
    Retrieves the Hatchet Session
    """
    global _hatchet_sessions
    return _hatchet_sessions.get(instance)

def set_hatchet_session(
    session: 'HatchetSession',
):
    """
    Sets the global Hatchet Session
    """
    global _hatchet_sessions
    _hatchet_sessions[session.instance] = session


def get_hatchet_client(**kwargs) -> 'HatchetClient':
    """
    Retrieves the Hatchet Client
    """
    global _hatchet_client
    if _hatchet_client is None:
        from lzl.api.hatchet.client import HatchetClient
        _hatchet_client = HatchetClient(**kwargs)
    return _hatchet_client

def set_hatchet_client(client: 'HatchetClient'):
    """
    Sets the global Hatchet Client
    """
    global _hatchet_client
    _hatchet_client = client

_patched_json = None

def patch_json():
    """
    Patches the json module to use our custom encoder and decoder
    """
    global _patched_json
    if _patched_json is not None: return
    import json
    json.loads = json_serializer.loads
    json.dumps = json_serializer.dumps
    _patched_json = True


def new_session(
    defaults: Union[ClientConfig, Dict[str, Any]] = None, 
    settings: Optional['HatchetSettings'] = None,
    instance: str = 'default',
    *opts_functions,
    namespace: Optional[str] = None,
    api_endpoint: Optional[str] = None,
    grpc_endpoint: Optional[str] = None,
    # include_instance_name: Optional[bool] = None,
    **kwargs,
) -> ClientImpl:
    """
    Create a new Hatchet Session instance
    """
    patch_json()
    if settings is None: settings = get_hatchet_settings()
    defaults = defaults or ClientConfig()
    settings.autologger.info(f'Initializing Hatchet Session: v{settings.version} (SDK Version: v{VERSION})', prefix = instance, colored = True)
    if not settings.in_k8s: set_ulimits(max_connections = 1024)
    if settings.in_k8s: os.environ['HATCHET_CLIENT_TLS_STRATEGY'] = 'none'
    
    config: ClientConfig = ConfigLoader(settings.config_lib_path or '.').load_client_config(defaults)

    # Here we do some overrides
    settings.configure_session_endpoints(
        config,
        instance = instance,
        api_endpoint = api_endpoint,
        grpc_endpoint = grpc_endpoint,
        **kwargs,
    )
    if not namespace and os.getenv(f'HATCHET_NAMESPACE_{instance.upper()}'):
        namespace = os.getenv(f'HATCHET_NAMESPACE_{instance.upper()}')
    
    config.namespace = namespace or f'{settings.app_env.short_name}.'
    if not settings.temp_data.has_logged(f'hatchet_init_{instance}'):
        logger.info(f'{config.server_url} - {config.host_port} (Namespace: {config.namespace})', prefix = f'Hatchet Config: {instance}', colored = True)
    
    for opt_function in opts_functions:
        opt_function(config)
    
    return ClientImpl.from_config(config)
    # if config.tls_config is None: raise ValueError("TLS config is required")
    # if config.host_port is None: raise ValueError("Host and port are required")
    # # config.logger = settings.logger

    # conn: grpc.Channel = new_conn(config)
    # # Instantiate client implementations
    # event_client = new_event(conn, config)
    # admin_client = new_admin(config)
    # dispatcher_client = new_dispatcher(config)
    # rest_client = RestApi(config.server_url, config.token, config.tenant_id)
    # workflow_listener_client = None
    # return ClientImpl(
    #     event_client,
    #     admin_client,
    #     dispatcher_client,
    #     workflow_listener_client,
    #     rest_client,
    #     config,
    # )





from __future__ import annotations

"""
A More Memory Efficient Version of Simdjson as it reuses the parser.
"""

import json
import simdjson
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from simdjson import Parser

# Once it hits this number, we will recycle the parser
MAX_PARSER_COUNT_TO_RECYCLE: int = 50
_parser: Optional['Parser'] = None
_counter: int = 0

def get_parser() -> 'Parser':
    global _parser, _counter
    if _counter > MAX_PARSER_COUNT_TO_RECYCLE:
        del _parser
        _parser = None
        _counter = 0
    if _parser is None: 
        _parser = simdjson.Parser()
    _counter += 1
    return _parser

def set_max_parser_count_to_recycle(count: int):
    global MAX_PARSER_COUNT_TO_RECYCLE
    MAX_PARSER_COUNT_TO_RECYCLE = count

def load(fp, *, cls=None, object_hook=None, parse_float=None, parse_int=None,
         parse_constant=None, object_pairs_hook=None, **kwargs):
    """
    Same as the built-in json.load(), with the following exceptions:

        - All parse_* arguments are currently ignored. simdjson does not
          currently provide hooks for these, but it is planned.
        - object_pairs_hook is ignored.
        - cls is ignored.
    """
    # parser = Parser()
    parser = get_parser()
    return parser.parse(fp.read(), True)


def loads(s, *, cls=None, object_hook=None, parse_float=None, parse_int=None,
          parse_constant=None, object_pairs_hook=None, **kwargs):
    """
    Same as the built-in json.loads(), with the following exceptions:

        - All parse_* arguments are currently ignored. simdjson does not
          currently provide hooks for these, but it is planned.
        - object_pairs_hook is ignored.
        - cls is ignored.
    """
    # parser = Parser()
    parser = get_parser()
    return parser.parse(s, True)


dumps = json.dumps
dump = json.dump


from __future__ import annotations

"""
Oso Base Class
"""

import functools
import typing as t
from oso import Oso as BaseOso
from polar.exceptions import DuplicateClassAliasError, OsoError, PolarRuntimeError
from polar import Variable
from polar.partial import TypeConstraint
from lzl.pool import ThreadPool
from lzl.logging import logger
from lzl import load
from . import errors

if t.TYPE_CHECKING:
    from sqlalchemy.orm import DeclarativeBase, registry
    from sqlalchemy.orm import Session
    from sqlalchemy.ext.asyncio import AsyncSession
    import sqlalchemy.sql.expression

if load.TYPE_CHECKING:
    import sqlalchemy
    import sqlalchemy_oso.partial
else:
    sqlalchemy = load.LazyLoad("sqlalchemy")
    sqlalchemy_oso = load.LazyLoad("sqlalchemy_oso", install_missing=True)

_Action = str
_Request = t.TypeVar("_Request")
_Resource = t.TypeVar("_Resource")
_Actor = t.TypeVar("_Actor")
_ResourceCls = t.Type[_Resource]

class Oso(BaseOso):
    """The central object to manage application policy state, e.g.
    the policy data, and verify requests.

    >>> from lzl.ext.oso import Oso
    >>> Oso()
    <oso.oso.Oso object at 0x...>
    """

    _registered_: t.Dict[str, t.Dict[str, t.Union[_ResourceCls, t.Any]]] = {}

    def __init__(
        self,
        *,
        forbidden_error: t.Type[t.Union[errors.OsoException, BaseException]] = errors.ForbiddenError,
        not_found_error: t.Type[t.Union[errors.OsoException, BaseException]] = errors.NotFoundError,
        read_action: _Action = "read"
    ) -> None:
        """
        Create an Oso object.

        :param forbidden_error:
            Optionally override the error class that is raised when an action is
            unauthorized.
        :param not_found_error:
            Optionally override the error class that is raised by the
            ``authorize`` method when an action is unauthorized AND the actor
            does not have permission to ``"read"`` the resource (and thus should
            not know it exists).
        :param read_action:
            The action used by the ``authorize`` method to determine whether an
            authorization failure should raise a ``NotFoundError`` or a
            ``ForbiddenError``.
        """
        super().__init__(forbidden_error = forbidden_error, not_found_error = not_found_error, read_action = read_action)
        self._pooler = ThreadPool
        self.postinit_hook()
        self.post_init_register_predefined()
        
        if t.TYPE_CHECKING:
            self.forbidden_error: t.Type[errors.OsoException]
            self.not_found_error: t.Type[errors.OsoException]
    
    def postinit_hook(self):
        """
        Post Initialization Hook
        """
        pass

    def _get_object_name(self, obj: t.Union[_Actor, _Resource]) -> str:
        """
        Returns the name of the object
        """
        if hasattr(obj, "name"): return obj.name
        if hasattr(obj, "__tablename__"): return obj.__tablename__
        if hasattr(obj, "__name__"): return obj.__name__
        return obj.__class__.__name__ if hasattr(obj, "__class__") else str(obj)

    def _raise_not_found(
        self,
        actor: _Actor,
        action: _Action,
        resource: _Resource,
    ) -> None:
        """
        Raises a NotFoundError
        """
        raise self.not_found_error(
            resource = self._get_object_name(resource),
            action = action,
            actor = self._get_object_name(actor),
        )

    def _raise_forbidden(
        self,
        actor: _Actor,
        action: _Action,
        resource: _Resource,
    ) -> None:
        """
        Raises a ForbiddenError
        """
        return self.forbidden_error(
            resource = self._get_object_name(resource),
            action = action,
            actor = self._get_object_name(actor),
        )

    def authorize(
        self,
        actor: _Actor,
        action: _Action,
        resource: _Resource,
        *,
        check_read: bool = True
    ) -> None:
        """Ensure that ``actor`` is allowed to perform ``action`` on
        ``resource``.

        If the action is permitted with an ``allow`` rule in the policy, then
        this method returns ``None``. If the action is not permitted by the
        policy, this method will raise an error.

        The error raised by this method depends on whether the actor can perform
        the ``"read"`` action on the resource. If they cannot read the resource,
        then a ``NotFound`` error is raised. Otherwise, a ``ForbiddenError`` is
        raised.

        :param actor: The actor performing the request.
        :param action: The action the actor is attempting to perform.
        :param resource: The resource being accessed.

        :param check_read: If set to ``False``, a ``ForbiddenError`` is always
            thrown on authorization failures, regardless of whether the actor can
            read the resource. Default is ``True``.
        :type check_read: bool

        """
        if self.query_rule_once("allow", actor, action, resource):
            return
        if check_read and (
            action == self.read_action
            or not self.query_rule_once("allow", actor, self.read_action, resource)
        ):
            self._raise_not_found(actor, action, resource)
        self._raise_forbidden(actor, action, resource)
        

    def authorize_request(self, actor: _Actor, request: _Request) -> None:
        """Ensure that ``actor`` is allowed to send ``request`` to the server.

        Checks the ``allow_request`` rule of a policy.

        If the request is permitted with an ``allow_request`` rule in the
        policy, then this method returns ``None``. Otherwise, this method raises
        a ``ForbiddenError``.

        :param actor: The actor performing the request.
        :param request: An object representing the request that was sent by the
            actor.
        """
        if not self.query_rule_once("allow_request", actor, request):
            self._raise_forbidden(actor, "request", request)
    

    def authorize_field(
        self, actor: _Actor, action: _Action, resource: _Resource, field: str
    ) -> None:
        """Ensure that ``actor`` is allowed to perform ``action`` on a given
        ``resource``'s ``field``.

        If the action is permitted by an ``allow_field`` rule in the policy,
        then this method returns ``None``. If the action is not permitted by the
        policy, this method will raise a ``ForbiddenError``.

        :param actor: The actor performing the request.
        :param action: The action the actor is attempting to perform on the
            field.
        :param resource: The resource being accessed.
        :param field: The name of the field being accessed.
        """
        if not self.query_rule_once("allow_field", actor, action, resource, field):
            self._raise_forbidden(actor, action, resource)



    """
    Async Methods
    """
    
    async def ais_allowed(
        self, actor: _Actor, action: _Action, resource: _Resource
    ) -> bool:
        """Evaluate whether ``actor`` is allowed to perform ``action`` on ``resource``.

        Uses allow rules in the Polar policy to determine whether a request is
        permitted. ``actor`` and ``resource`` should be classes that have been
        registered with Polar using the :py:func:`register_class` function.

        :param actor: The actor performing the request.
        :param action: The action the actor is attempting to perform.
        :param resource: The resource being accessed.

        :return: ``True`` if the request is allowed, ``False`` otherwise.
        """
        return await self._pooler.arun(self.is_allowed, actor, action, resource)

    async def aget_allowed_actions(
        self, actor: _Actor, resource: _Resource, allow_wildcard: bool = False
    ) -> t.List[t.Any]:
        """Determine the actions ``actor`` is allowed to take on ``resource``.
        """
        return list(await self.aauthorized_actions(actor, resource, allow_wildcard = allow_wildcard))

    async def aauthorize(
        self,
        actor: _Actor,
        action: _Action,
        resource: _Resource,
        *,
        check_read: bool = True
    ) -> None:
        """Ensure that ``actor`` is allowed to perform ``action`` on
        ``resource``.

        If the action is permitted with an ``allow`` rule in the policy, then
        this method returns ``None``. If the action is not permitted by the
        policy, this method will raise an error.

        The error raised by this method depends on whether the actor can perform
        the ``"read"`` action on the resource. If they cannot read the resource,
        then a ``NotFound`` error is raised. Otherwise, a ``ForbiddenError`` is
        raised.

        :param actor: The actor performing the request.
        :param action: The action the actor is attempting to perform.
        :param resource: The resource being accessed.

        :param check_read: If set to ``False``, a ``ForbiddenError`` is always
            thrown on authorization failures, regardless of whether the actor can
            read the resource. Default is ``True``.
        :type check_read: bool

        """
        return await self._pooler.arun(self.authorize, actor, action, resource, check_read = check_read)

    async def aauthorize_request(self, actor: _Actor, request: _Request) -> None:
        """Ensure that ``actor`` is allowed to send ``request`` to the server.

        Checks the ``allow_request`` rule of a policy.

        If the request is permitted with an ``allow_request`` rule in the
        policy, then this method returns ``None``. Otherwise, this method raises
        a ``ForbiddenError``.

        :param actor: The actor performing the request.
        :param request: An object representing the request that was sent by the
            actor.
        """
        return await self._pooler.arun(self.authorize_request, actor, request)
    
    async def aauthorized_actions(
        self, actor: _Actor, resource: _Resource, allow_wildcard: bool = False
    ) -> t.Set[t.Any]:
        """Determine the actions ``actor`` is allowed to take on ``resource``.

        Collects all actions allowed by allow rules in the Polar policy for the
        given combination of actor and resource.

        Identical to ``Oso.get_allowed_actions``.

        :param actor: The actor for whom to collect allowed actions

        :param resource: The resource being accessed

        :param allow_wildcard: Flag to determine behavior if the policy
            contains an "unconstrained" action that could represent any action:
            ``allow(_actor, _action, _resource)``. If ``True``, the method will
            return ``["*"]``, if ``False`` (the default), the method will raise
            an exception.

        :type allow_wildcard: bool

        :return: A set containing all allowed actions.
        """
        return await self._pooler.arun(self.authorized_actions, actor, resource, allow_wildcard = allow_wildcard)

    async def aauthorize_field(
        self, actor: _Actor, action: _Action, resource: _Resource, field: str
    ) -> None:
        """Ensure that ``actor`` is allowed to perform ``action`` on a given
        ``resource``'s ``field``.

        If the action is permitted by an ``allow_field`` rule in the policy,
        then this method returns ``None``. If the action is not permitted by the
        policy, this method will raise a ``ForbiddenError``.

        :param actor: The actor performing the request.
        :param action: The action the actor is attempting to perform on the
            field.
        :param resource: The resource being accessed.
        :param field: The name of the field being accessed.
        """
        return await self._pooler.arun(self.authorize_field, actor, action, resource, field)
    

    async def aauthorized_fields(
        self,
        actor: _Actor,
        action: _Action,
        resource: _Resource,
        allow_wildcard: bool = False,
    ) -> t.Set[t.Any]:
        """Determine the fields of ``resource`` on which ``actor`` is allowed to
        perform  ``action``.

        Uses ``allow_field`` rules in the policy to find all allowed fields.

        :param actor: The actor for whom to collect allowed fields.
        :param action: The action being taken on the fields.
        :param resource: The resource being accessed.

        :param allow_wildcard: Flag to determine behavior if the policy \
            includes a wildcard field. E.g., a rule allowing any field: \
            ``allow_field(_actor, _action, _resource, _field)``. If ``True``, the \
            method will return ``["*"]``, if ``False``, the method will raise an \
            exception.

        :type allow_wildcard: bool

        :return: A set containing all allowed fields.
        """
        return await self._pooler.arun(self.authorized_fields, actor, action, resource, allow_wildcard = allow_wildcard)
    

    async def aauthorized_query(
        self, actor: _Actor, action: _Action, resource_cls: _ResourceCls
    ):
        """Create a query for resources of type ``resource_cls``
        that ``actor`` is allowed to perform ``action`` on. The
        query is built by using the ``build_query`` and ``combine_query``
        functions registered for the ``resource_cls``.

        :param actor: The actor for whom to collect allowed resources.
        :param action: The action that user wants to perform.
        :param resource_cls: The type of the resources.

        :return: A query to fetch the resources,
        """
        return await self._pooler.arun(self.authorized_query, actor, action, resource_cls)
    

    async def aauthorized_resources(
        self, actor: _Actor, action: _Action, resource_cls: _ResourceCls
    ) -> t.List[t.Any]:
        """Determine the resources of type ``resource_cls`` that ``actor``
        is allowed to perform ``action`` on.

        :param actor: The actor for whom to collect allowed resources.
        :param action: The action that user wants to perform.
        :param resource_cls: The type of the resources.

        :return: The requested resources.
        """
        return await self._pooler.arun(self.authorized_resources, actor, action, resource_cls)
    
    @staticmethod
    def get_sa_field_type(model: t.Type['DeclarativeBase'], field: str) -> t.Type['DeclarativeBase']:
        """
        Returns the type of the field
        """
        try:
            field = getattr(model, field)
        except AttributeError as e:
            raise PolarRuntimeError(f"Cannot get property {field} on {model}.") from e

        try:
            return field.entity.class_
        except AttributeError as e:
            raise PolarRuntimeError(
                f"Cannot determine type of {field} on {model}."
            ) from e

    @staticmethod
    def iterate_sa_model_classes(base_or_registry: t.Union[t.Type['DeclarativeBase'], 'registry']):
        """
        Return an iterator of model classes that descend from a declarative base
        (SQLAlchemy 1.3 or 1.4) or exist in a registry (SQLAlchemy 1.4).
        """
        try:  # 1.3 declarative base.
            # TODO (dhatch): Not sure this is legit b/c it uses an internal interface?
            models = base_or_registry._decl_class_registry.items()
            for name, model in models:
                if name != "_sa_module_registry":
                    if isinstance(
                        model, sqlalchemy.ext.declarative.clsregistry._MultipleClassMarker
                    ):
                        for model_ref in model.contents:
                            yield model_ref()
                    else:
                        yield model
        except AttributeError:
            try:  # 1.4 declarative base.
                mappers = base_or_registry.registry.mappers
            except AttributeError:  # 1.4 registry.
                mappers = base_or_registry.mappers
            yield from {mapper.class_ for mapper in mappers}

    @staticmethod
    def default_polar_model_name(model: type) -> str:
        """
        Return polar class name for SQLAlchemy model.
        """
        return model.__name__

    def register_sql_models(
        self, 
        *base_or_registry: t.Type['DeclarativeBase'],
    ):
        """
        Register all models in registry (SQLAlchemy 1.4) or declarative base
        class (1.3 and 1.4) ``base_or_registry`` with Oso as classes.
        """
        for registry in base_or_registry:
            for model in self.iterate_sa_model_classes(registry):
                if model in self.host.types:
                    # skip models that were manually registered
                    continue
                try:
                    self.register_class(model, name = self.default_polar_model_name(model))
                except DuplicateClassAliasError as e:
                    raise OsoError(
                        "Attempted to register two classes with the same name when automatically registering SQLAlchemy models\n"
                        "To fix this, try manually registering the new class. E.g.\n"
                        '  oso.register_class(MyModel, name="models::MyModel")\n'
                        "  register_models(oso, Base)\n"
                    ) from e

    def authorize_sa_model(
        self,
        actor: _Actor,
        action: _Action,
        session: t.Union['Session', 'AsyncSession'],  # sourcery skip: avoid-builtin-shadow
        model: t.Type['DeclarativeBase'],
    ):    # sourcery skip: avoid-builtin-shadow
        """Return SQLAlchemy expression that applies the policy to ``model``.

        Executing this query will return only authorized objects. If the request is
        not authorized, a query that always contains no result will be returned.

        :param oso: The oso class to use for evaluating the policy.
        :param actor: The actor to authorize.
        :param action: The action to authorize.

        :param session: The SQLAlchemy session.
        :param model: The model to authorize, must be a SQLAlchemy model or alias.
        """
        self.host.get_field = self.get_sa_field_type
        try:
            mapped_class = sqlalchemy.inspect(model, raiseerr=True).class_
        except AttributeError:
            raise TypeError(f"Expected a model; received: {model}")

        user_type = self.host.types.get(mapped_class)
        model_name = (
            user_type.name or self.default_polar_model_name(mapped_class)
            if user_type
            else self.default_polar_model_name(mapped_class)
        )
        resource = Variable("resource")
        constraint = TypeConstraint(resource, model_name)
        results = self.query_rule(
            "allow",
            actor,
            action,
            resource,
            bindings={resource: constraint},
            accept_expression=True,
        )
        combined_filter = None
        has_result = False
        for result in results:
            has_result = True
            resource_partial = result["bindings"]["resource"]
            if isinstance(resource_partial, model):
                def f(pk):
                    return getattr(model, pk) == getattr(resource_partial, pk)

                filters = [f(pk.name) for pk in sqlalchemy.inspect(model).primary_key]
                filter = functools.reduce(lambda a, b: a & b, filters)
            else:
                filter = sqlalchemy_oso.partial.partial_to_filter(
                    resource_partial, 
                    session, 
                    model, 
                    get_model = self.get_class
                )
            if combined_filter is None: combined_filter = filter
            else: combined_filter = combined_filter | filter
        return combined_filter if has_result else sqlalchemy.sql.expression.false()

    def register_classes(self, *classes: t.Type[_Actor]):
        """
        Registers the classes
        """
        for ct in classes:
            self.register_class(ct)

    def post_init_register_predefined(self, **kwargs):
        """
        Registers the predefined objects
        """
        if not self._registered_: return
        for name, config in self._registered_.items():
            new_cls = config.pop('cls')
            if isinstance(new_cls, str): new_cls = load.lazy_import(new_cls)
            new = {
                'name': name,
                **config,
            }
            self.register_class(new_cls, **new)

    @classmethod
    def _preregister_class(cls, model: _ResourceCls, name: t.Optional[str] = None, fields: t.Optional[t.List[str]] = None):
        """
        Preregisters a model
        """
        if name is None: name = cls.default_polar_model_name(model)
        cls._registered_[name] = {
            'cls': model,
            'fields': fields,
        }

    @classmethod
    def preregister_class(cls, name: t.Optional[str] = None, fields: t.Optional[t.List[str]] = None):
        """
        Preregisters the models
        """
        def wrapper(model: _ResourceCls):
            cls._preregister_class(model, name = name, fields = fields)
            return model
        return wrapper

    @classmethod
    def _preregister_lazy(cls, model_source: str, name: t.Optional[str] = None, fields: t.Optional[t.List[str]] = None):
        """
        Preregisters a model
        """
        if name is None: name = model_source.rsplit('.', 1)[-1]
        cls._registered_[name] = {
            'cls': model_source,
            'fields': fields,
        }

    @classmethod
    def preregister_lazy(cls, model_source: t.Union[str, t.List[str], t.Dict[str, t.Any]], name: t.Optional[str] = None, fields: t.Optional[t.List[str]] = None):
        """
        Lazily Preregisters the models
        """
        if isinstance(model_source, str):
            return cls._preregister_lazy(model_source, name = name, fields = fields)
        if isinstance(model_source, dict):
            for k, v in model_source.items():
                cls._preregister_lazy(v, name = k, fields = fields)
            return
        if isinstance(model_source, list):
            for v in model_source:
                cls._preregister_lazy(v, name = name, fields = fields)
            return
        raise ValueError(f'Invalid model_source: {model_source}')

    @classmethod
    def preregister(cls, model_or_source: t.Optional[t.Union[str, t.List[str], t.Dict[str, t.Any], _ResourceCls]] = None, name: t.Optional[str] = None, fields: t.Optional[t.List[str]] = None):
        """
        PreRegisters the models
        """
        if model_or_source is None: return cls.preregister_class(name = name, fields = fields)
        if isinstance(model_or_source, (str, list, dict)):
            return cls.preregister_lazy(model_or_source, name = name, fields = fields)
        return cls._preregister_class(model_or_source, name = name, fields = fields)
    



from __future__ import annotations

from ..types.config import BaseOAuth2Config, eproperty
from ..types.client import BaseOAuth2Client, jwt
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING, List, Type, Annotated
from pydantic import Field
from lzo.utils import create_unique_id

_default_scopes = [
    'openid',
    'https://www.googleapis.com/auth/userinfo.email',
    'https://www.googleapis.com/auth/userinfo.profile',
]

class GoogleOAuth2Config(BaseOAuth2Config):
    """
    Google OAuth2 Config
    """
    name: Optional[str] = 'google'
    issuer_url: Optional[str] = None
    jwt_issuer_url: Optional[str] = 'https://accounts.google.com'
    scopes: Optional[List[str]] = Field(default = _default_scopes)
    jwt_url: Optional[str] = 'https://www.googleapis.com/oauth2/v3/certs'

    authorize_disabled: Optional[bool] = True
    
    class Config(BaseOAuth2Config.Config):
        env_prefix = "GOOGLE_"

    @eproperty
    def authorization_endpoint(self) -> str:
        """
        Returns the authorization endpoint
        """
        return 'https://accounts.google.com/o/oauth2/v2/auth'
    
    @eproperty
    def token_endpoint(self) -> str:
        """
        Returns the token endpoint
        """
        return 'https://oauth2.googleapis.com/token'
    
    @eproperty
    def token_info_endpoint(self) -> str:
        """
        Returns the token info endpoint
        """
        return 'https://oauth2.googleapis.com/tokeninfo'
    
    @eproperty
    def user_info_endpoint(self) -> str:
        """
        Returns the user info endpoint
        """
        return 'https://openidconnect.googleapis.com/v1/userinfo'

    def get_scopes(
        self,
        scopes: Optional[List[str]] = None,
        sep: Optional[str] = ' ',
    ) -> str:
        """
        Returns the scopes
        """
        if not scopes: return self.scope
        for n, scope in enumerate(scopes):
            if scope == 'openid': continue
            # This deals with `userinfo.email` and `userinfo.profile`
            if not scope.startswith('https://'):
                scopes[n] = f'https://www.googleapis.com/auth/{scope}'
        return sep.join(scopes)

    def get_auth_params(
        self,
        redirect_url: str,
        scope: Optional[str] = None,
        state: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the auth params
        """
        scope = scope or self.get_scopes()
        state = state or create_unique_id(alph_only=True)
        return {
            'redirect_uri': redirect_url,
            'client_id': self.client_id,
            'scope': scope,
            'access_type': 'offline',
            'prompt': 'consent',
            'response_type': 'code',
            'state': state,
            'include_granted_scopes': 'true',
        }
    
    def get_fetch_token_params(
        self,
        code: str,
        redirect_url: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the fetch token params
        """
        return {
            'code': code,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'redirect_uri': redirect_url,
            'grant_type': 'authorization_code',
        }
    
    def get_refresh_token_params(
        self,
        refresh_token: str,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the refresh token params
        """
        return {
            'refresh_token': refresh_token,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': 'refresh_token',
        }


class GoogleOAuth2Client(BaseOAuth2Client):
    """
    Google Auth Client
    """
    name: Optional[str] = 'google'

    if TYPE_CHECKING:
        config: GoogleOAuth2Config
        
    
    def parse_jwt_token(
        self,
        jwt_token: str,
        kind: Optional[str] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Parses the jwt token
        """
        # Posssible bug with the jwt?
        if kind == 'access_token': return None
        return jwt.decode(
            jwt_token,
            "",
            # signing_key.key, 
            algorithms=['RS256'],
            issuer = self.config.issuer_url,
            options = {
                'verify_signature': False,
                'verify_exp': True,
                'verify_aud': False,
            }
        )
        # return super().parse_jwt_token(jwt_token, kind = kind)

    
    def parse_authorization_token(
        self,
        authorization: str,
        audiences: Optional[List[str]] = None,
        verify_exp: bool = True,
    ) -> Optional[Dict[str, Any]]:
        """
        Parses the authorization token
        """
        # Posssible bug with the jwt?
        return jwt.decode(
            authorization, 
            "", 
            algorithms = ['RS256'],
            audience = audiences,
            issuer = self.config.jwt_issuer_url,
            options={
                'verify_signature': False,
                'verify_exp': verify_exp,
                'verify_aud': bool(audiences),
            }
        )
    
    @eproperty
    def config_class(self) -> Type['GoogleOAuth2Config']:
        """
        Returns the Config Class
        """
        return GoogleOAuth2Config
    

    async def afetch_user_info(
        self,
        token: str,
        token_payload: Dict[str, Any],
        **kwargs,
    ):
        """
        Fetches the user info
        """

    async def afetch_access_token_info(
        self,
        token: str,
        token_payload: Dict[str, Any],
        **kwargs,
    ):
        """
        Fetches the access token info
        """
        pass


    async def abuild_oauth_token(
        self,
        token: str,
        token_payload: Dict[str, Any],
        **kwargs,
    ) -> Optional[Dict[str, Any]]:
        """
        Builds the oauth token
        """
        # we'll try to determine the token type based on the prefix
        token_data = {}
        if token.startswith('ya29.'):
            token_data['access_token'] = {**(await self.afetch_access_token_info(token, token_payload)), 'value': token}
        else:
            token_data['id_token'] = {**self.parse_jwt_token(token, kind = 'id_token'), 'value': token}
        return token_data





        


from .base import BaseCompression
from ._gzip import GzipCompression
from ._lz4 import Lz4Compression, _lz4_available
from ._zlib import ZlibCompression
from ._zstd import ZstdCompression, _zstd_available
from typing import Any, Dict, Optional, Union, Type


CompressionT = Union[GzipCompression, Lz4Compression, ZlibCompression, ZstdCompression, BaseCompression]


DEFAULT_COMPRESSION = (
    "zstd" if _zstd_available else \
    "lz4" if _lz4_available else \
    "zlib"
)

def get_default_compression() -> str:
    """
    Returns the default serializer
    """
    return DEFAULT_COMPRESSION

def set_default_compression(
    compression: str,
) -> None:
    """
    Sets the default compression

    :param compression: The compression to use
    """
    global DEFAULT_COMPRESSION
    DEFAULT_COMPRESSION = compression

_initialized_compressors: Dict[str, CompressionT] = {}

def get_compression(
    compression_type: Optional[str] = None,
    compression_level: Optional[int] = None,
    **kwargs
) -> CompressionT:
    """
    Returns a Compression
    """
    from lzo.utils.hashing import create_hash_from_args_and_kwargs
    if compression_type == 'auto': compression_type = None
    compression_type = compression_type or get_default_compression()
    comp_hash = create_hash_from_args_and_kwargs(compression_type, compression_level = compression_level, **kwargs)
    if comp_hash in _initialized_compressors:
        return _initialized_compressors[comp_hash]
    if compression_type == "gzip":
        new = GzipCompression(compression_level = compression_level, **kwargs)
    elif compression_type == "lz4":
        new = Lz4Compression(compression_level = compression_level, **kwargs)
    elif compression_type == "zlib":
        new = ZlibCompression(compression_level = compression_level, **kwargs)
    elif compression_type == "zstd":
        new = ZstdCompression(compression_level = compression_level, **kwargs)
    else: raise ValueError(f"Invalid Compression Type: {compression_type}")
    _initialized_compressors[comp_hash] = new
    return new